#
# Copyright 2016 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#

import pytest

import numpy as np

from ..interpolation import SteffenInterpolator


def test_extrapolation_raise_by_default():
    xi = np.arange(0, 10, 0.5)
    yi = 3 * xi

    stfi = SteffenInterpolator(xi, yi)
    with pytest.raises(ValueError):
        stfi(200)

    stfi1 = SteffenInterpolator(xi, yi, extrapolate='raise')
    with pytest.raises(ValueError):
        stfi1(200)


def test_extrapolation_check_unknown_mode():
    xi = np.arange(0, 10, 0.5)
    yi = 3 * xi

    with pytest.raises(ValueError):
        SteffenInterpolator(xi, yi, extrapolate='kisjd')


def test_extrapolation_check_known_mode():
    xi = np.arange(0, 10, 0.5)
    yi = 3 * xi

    m = SteffenInterpolator(xi, yi, extrapolate='raise')
    assert isinstance(m, SteffenInterpolator)

    m = SteffenInterpolator(xi, yi, extrapolate='zeros')
    assert isinstance(m, SteffenInterpolator)

    m = SteffenInterpolator(xi, yi, extrapolate='const')
    assert isinstance(m, SteffenInterpolator)

    m = SteffenInterpolator(xi, yi, extrapolate='border')
    assert isinstance(m, SteffenInterpolator)

    m = SteffenInterpolator(xi, yi, extrapolate='extrapolate')
    assert isinstance(m, SteffenInterpolator)


def test_interpolation():

    def function(x):
        return x * (12*np.sin(x)-3)-0.3*x**2*(8*np.cos(3.1*x)-3) / (x+23.4)

    xi = np.arange(0, 6.5, 0.5)
    yi = function(xi)

    sti = SteffenInterpolator(xi, yi)
    xnew = np.arange(0, 6.5, 0.5)
    assert np.allclose(sti(xnew), yi)

    xnew = np.linspace(1.2, 1.4, 10)

    expected = np.array([9.94082135,  10.2539533,  10.56475383,  10.87191933, 11.1741462,
                         11.47013086, 11.7585697, 12.03815913, 12.30759555,  12.56557537])
    result = sti(xnew)

    assert np.allclose(result, expected)


def test_extrapolation_zeros():

    def function(x):
        return x * (12*np.sin(x)-3)-0.3*x**2*(8*np.cos(3.1*x)-3) / (x+23.4)

    xi = np.arange(0, 6.0)
    yi = 1.0 + 3 * xi

    sti = SteffenInterpolator(xi, yi, extrapolate='zeros')

    xnew = [6.0, 7.0]
    result = sti(xnew)
    expected = [0.0, 0.0]
    assert np.allclose(result, expected)

    xnew = [-1.0, -0.1]
    result = sti(xnew)
    expected = [0.0, 0.0]
    assert np.allclose(result, expected)


def test_extrapolation_const():

    xi = np.arange(0, 6.0)
    yi = 1.0 + 3 * xi

    fill = 12.0
    sti = SteffenInterpolator(xi, yi, extrapolate='const', fill_value=fill)

    xnew = [6.0, 7.0]
    result = sti(xnew)
    expected = [fill, fill]
    assert np.allclose(result, expected)

    xnew = [-1.0, -0.1]
    result = sti(xnew)
    expected = [fill, fill]
    assert np.allclose(result, expected)


def test_extrapolation_border():

    xi = np.arange(0, 6.0)
    yi = 1.0 + 3 * xi

    sti = SteffenInterpolator(xi, yi, extrapolate='border')

    xnew = [6.0, 7.0]
    result = sti(xnew)
    expected = [yi[-1], yi[-1]]
    assert np.allclose(result, expected)

    xnew = [-1.0, -0.1]
    result = sti(xnew)
    expected = [yi[0], yi[0]]
    assert np.allclose(result, expected)


def test_extrapolation_extrapolate():

    xi = np.arange(0, 6.0)
    yi = 1.0 + 3 * xi

    sti = SteffenInterpolator(xi, yi, extrapolate='extrapolate')

    xnew = [6.0, 7.0]
    result = sti(xnew)
    expected = [7, -32.0]
    assert np.allclose(result, expected)

    xnew = [-1.0, -0.1]
    result = sti(xnew)
    expected = [10.0, 1.063]
    assert np.allclose(result, expected)


def test_conditions_in_borders():

    def function(x):
        return x * (12*np.sin(x)-3)-0.3*x**2*(8*np.cos(3.1*x)-3) / (x+23.4)

    xi = np.arange(0, 6.0)
    yi = 1.0 + 3 * xi

    with pytest.raises(ValueError):
        SteffenInterpolator(xi, yi, yp_0=200)

    with pytest.raises(ValueError):
        SteffenInterpolator(xi, yi, yp_0=200)

    with pytest.raises(ValueError):
        SteffenInterpolator(xi, yi, yp_N=200)

    with pytest.raises(ValueError):
        SteffenInterpolator(xi, yi, yp_N=-200)





import numpy
from numina.array.bpm import process_bpm_median


def test_process_bpm():

    data = numpy.zeros((10, 10), dtype='float32') + 3.0
    mask = numpy.zeros((10, 10), dtype='int32')
    mask[3, 3] = 1

    result1 = process_bpm_median(data, mask)

    assert result1[3, 3] == 3.0

    result2, subs2 = process_bpm_median(data, mask, subs=True)

    assert result2[3, 3] == 3.0
    assert subs2.min() == 1


def test_process_bpm_large_hole():

    data = numpy.zeros((100, 100), dtype='float32') + 3.0
    mask = numpy.zeros((100, 100), dtype='int32')
    mask[30:40, 30:40] = 1

    fill = 0.1
    result, subs = process_bpm_median(data, mask, fill=fill, subs=True)

    assert result[35, 35] == fill
    assert subs.sum() == 100 * 100 - 36

    result1, subs1 = process_bpm_median(
        data, mask, reuse_values=True, subs=True)

    assert result1[35, 35] == 3.0
    assert subs1.sum() == 100 * 100

    result2 = process_bpm_median(data, mask, reuse_values=True, subs=False)

    assert result2[35, 35] == 3.0



#
# Copyright 2015-2018 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#

import numpy

from ..hwdevice import HWDevice


class VirtualDetector:
    """Each of the channels."""

    def __init__(self, base, geom, directfun, readpars):

        self.base = base
        self.trim, self.pcol, self.ocol, self.orow = geom

        self.direcfun = directfun

        self.readpars = readpars

    def readout_in_buffer(self, elec, final):

        final[self.trim] = self.direcfun(elec[self.base])

        final[self.trim] = final[self.trim] / self.readpars.gain

        # We could use different RON and BIAS in each section
        for section in [self.trim, self.pcol, self.ocol, self.orow]:
            final[section] = self.readpars.bias + numpy.random.normal(
                final[section], self.readpars.ron
            )

        return final


class DetectorBase(HWDevice):
    def __init__(self, name, shape, qe=1.0, qe_wl=None, dark=0.0):

        super().__init__(name)

        self.dshape = shape
        self.pixscale = 15.0e-3

        self._det = numpy.zeros(shape, dtype="float64")

        self.qe = qe

        if qe_wl is None:
            self._qe_wl = 1.0
        else:
            self._qe_wl = qe_wl

        self.dark = dark
        # Exposure time since last reset
        self._time_last = 0.0

    def qe_wl(self, wl):
        """QE per wavelength."""
        return self._qe_wl.response(wl)

    def expose(self, source=0.0, time=0.0):
        self._time_last = time
        self._det += (self.qe * source + self.dark) * time

    def reset(self):
        """Reset the detector."""
        self._det[:] = 0.0

    def saturate(self, x):
        return x

    def simulate_poisson_variate(self):
        elec_mean = self._det
        elec = numpy.random.poisson(elec_mean)
        return elec

    def pre_readout(self, elec_pre):
        return elec_pre

    def base_readout(self, elec_f):
        return elec_f

    def post_readout(self, adu_r):
        adu_p = numpy.clip(adu_r, 0, 2**16 - 1)
        return adu_p.astype("uint16")

    def clean_up(self):
        self.reset()

    def readout(self):
        """Readout the detector."""

        elec = self.simulate_poisson_variate()

        elec_pre = self.saturate(elec)

        elec_f = self.pre_readout(elec_pre)

        adu_r = self.base_readout(elec_f)

        adu_p = self.post_readout(adu_r)

        self.clean_up()

        return adu_p



#
# Copyright 2015-2023 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#

"""Robust fits"""

import numpy


def fit_theil_sen(x, y):
    """Compute a robust linear fit using the Theil-Sen method.

    See http://en.wikipedia.org/wiki/Theil%E2%80%93Sen_estimator for details.
    This function "pairs up sample points by the rank of their x-coordinates
    (the point with the smallest coordinate being paired with the first point
    above the median coordinate, etc.) and computes the median of the slopes of
    the lines determined by these pairs of points".

    Parameters
    ----------
    x : array_like, shape (M,)
        X coordinate array. This array must be sorted.
    y : array_like, shape (M,) or (M,K)
        Y coordinate array. If the array is two dimensional, each column of
        the array is independently fitted sharing the same x-coordinates. In
        this last case, the returned intercepts and slopes are also 1d numpy
        arrays. This array must be properly sorted to match the same order
        as the input X coordinate array.

    Returns
    -------
    coef : ndarray, shape (2,) or (2, K)
           Intercept and slope of the linear fit. If y was 2-D, the
           coefficients in column k of coef represent the linear fit
           to the data in y's k-th column.

    Raises
    ------
    ValueError:
        If the number of points to fit is < 5

    """

    xx = numpy.asarray(x)
    y1 = numpy.asarray(y)
    n = len(xx)
    if n < 5:
        raise ValueError('Number of points < 5')

    if xx.ndim != 1:
        raise ValueError('Input arrays have unexpected dimensions')

    diff = xx[1:] - x[:-1]
    if not numpy.all(diff > 0):
        raise ValueError('The input arrays must be sorted')

    if y1.ndim == 1:
        if len(y1) != n:
            raise ValueError('X and Y arrays have different sizes')
        yy = y1[numpy.newaxis, :]
    elif y1.ndim == 2:
        if n != y1.shape[0]:
            raise ValueError(
                'Y-array size in the fitting direction is different to the X-array size')
        yy = y1.T
    else:
        raise ValueError('Input arrays have unexpected dimensions')

    nmed = n // 2
    iextra = nmed if (n % 2) == 0 else nmed + 1

    deltx = xx[iextra:] - xx[:nmed]
    delty = yy[:, iextra:] - yy[:, :nmed]
    allslopes = delty / deltx
    slopes = numpy.median(allslopes, axis=1)
    allinters = yy - slopes[:, numpy.newaxis] * x
    inters = numpy.median(allinters, axis=1)

    coeff = numpy.array([inters, slopes])
    return numpy.squeeze(coeff)



#
# Copyright 2008-2018 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#


"""Extra logging handlers for the numina logging system."""


import logging

from astropy.io import fits


class FITSHistoryHandler(logging.Handler):
    """Logging handler using HISTORY FITS cards"""

    def __init__(self, header):
        logging.Handler.__init__(self)
        self.header = header

    def emit(self, record):
        msg = self.format(record)
        self.header.add_history(msg)


def log_to_history(logger, name):
    """Decorate function, adding a logger handler stored in FITS."""

    def log_to_history_decorator(method):

        def l2h_method(self, ri):
            history_header = fits.Header()

            fh = FITSHistoryHandler(history_header)
            fh.setLevel(logging.INFO)
            logger.addHandler(fh)

            try:
                result = method(self, ri)
                field = getattr(result, name, None)
                if field:
                    with field.open() as hdulist:
                        hdr = hdulist[0].header
                        hdr.extend(history_header.cards)
                return result
            finally:
                logger.removeHandler(fh)
        return l2h_method

    return log_to_history_decorator



#
# Copyright 2015-2018 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#


"""Utilities for DAL"""


def tags_are_valid(subset, superset):
    """Validate tags"""
    for key, val in subset.items():
        if key in superset and superset[key] != val:
            return False
    return True



#
# Copyright 2014-2023 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#

import numpy as np
from numpy.testing import assert_allclose

from ... import modeling


def test_EnclosedGaussian():

    A = 1234.02
    sigma = 3.1

    model = modeling.EnclosedGaussian(A, sigma)

    rad = np.arange(0, 15, 1.0)
    Fr_ref = np.array([0., 62.56332417, 231.85707671, 461.41325153,
                       697.25186078, 897.9551982, 1044.40731257,
                       1137.61050118, 1189.84461703, 1215.77898974,
                       1227.232255, 1231.74380107, 1233.33213537,
                       1233.83267152, 1233.97402596])

    Fr = model(rad)

    assert_allclose(Fr, Fr_ref, rtol=0, atol=1e-6)


def test_GaussBox():

    a = 34.5
    s = 1.5
    xp = np.arange(-5, 6, 1.0)
    yp_ref = np.array([0.00122703, 0.00846543, 0.03797502, 0.1108649,
                       0.21078609, 0.26111732, 0.21078609, 0.1108649,
                       0.03797502, 0.00846543, 0.00122703]) * a

    model = modeling.GaussBox(a, 0.0, s, 0.5)
    yp = model(xp)

    assert_allclose(yp, yp_ref, rtol=0, atol=1e-6)


def test_sum_of_gaussians():
    N = 7
    amps = np.array([1e2, 1.01e2, 7.015e2, 1.417e2, 2.021e2, 1.022e2, 1.025e1])
    centers = [7.63, 14.24, 21.35, 28.1, 34.7, 41.9, 49.2]
    sigs = [1.48, 1.51, 1.46, 1.48, 1.51, 1.46, 1.50]
    back = 120.0

    xl = np.arange(0, 9 * N, dtype='float')
    yl = np.zeros_like(xl)
    for a, l, s in zip(amps[:N], centers[:N], sigs[:N]):
        z = (xl - l) / s
        yl += a * np.exp(-0.5 * z ** 2)

    yl += back

    init = {}
    for i in range(N):
        init[f'amplitude_{i:d}'] = amps[i]
        init[f'center_{i:d}'] = centers[i]
        init[f'stddev_{i:d}'] = sigs[i]
    init['background'] = back

    KS = modeling.sum_of_gaussian_factory(N)
    model_gs = KS(**init)
    yp = model_gs(xl)

    assert_allclose(yl, yp, rtol=0, atol=1e-6)




from .seffect import side_effect, record_call, FuncCall


def test_side_effect_1():

    somevals = []

    def sidefunc(pos):
        somevals.append(pos)

    @side_effect(sidefunc)
    def testfunc(pos):
        return pos

    assert testfunc(1) == 1

    assert somevals == [1]


def test_side_effect_except():

    somevals = []

    def sidefunc(pos):
        a = 1 / 0
        somevals.append(a)

    @side_effect(sidefunc)
    def testfunc(pos):
        return pos

    assert testfunc(1) == 1
    # Exception is ignored, no side effect
    assert somevals == []


def test_record_call():

    @record_call
    def testfunc(pos):
        return pos

    assert testfunc.side_effect.args == ()
    assert testfunc.side_effect.kwargs == {}
    assert testfunc.side_effect.called is False

    assert testfunc(1) == 1
    assert isinstance(testfunc.side_effect, FuncCall)

    assert testfunc.side_effect.args == (1,)
    assert testfunc.side_effect.kwargs == {}
    assert testfunc.side_effect.called is True

    testfunc.side_effect.clear()

    assert testfunc.side_effect.args == ()
    assert testfunc.side_effect.kwargs == {}
    assert testfunc.side_effect.called is False




import pytest

import numina.core.pipelineload as loader


@pytest.fixture(scope='module')
def drptest():
    return loader.drp_load('numina.drps.tests', 'drptest1.yaml')


def test_mode_search(drptest):

    ll = drptest.search_mode_provides("MasterBias")

    assert ll.name == 'MasterBias'
    assert ll.mode == drptest.modes[ll.mode].key
    assert ll.field == 'master_bias'

    ll = drptest.search_mode_provides("MasterDark")

    assert ll.name == 'MasterDark'
    assert ll.mode == drptest.modes[ll.mode].key
    assert ll.field == 'master_dark'


def test_mode_query(drptest):

    ll = drptest.query_provides("MasterBias")

    assert ll.name == 'MasterBias'
    assert ll.mode == drptest.modes[ll.mode].key
    assert ll.field == 'master_bias'

    with pytest.raises(ValueError):
        drptest.query_provides("MasterDark")

    ll = drptest.query_provides("MasterDark", search=True)

    assert ll.name == 'MasterDark'
    assert ll.mode == drptest.modes[ll.mode].key
    assert ll.field == 'master_dark'



#
# Copyright 2008-2023 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#

from copy import deepcopy
import numbers
import numpy as np

from numpy.polynomial import Polynomial

from numina.array.display.polfit_residuals import polfit_residuals
from numina.array.display.polfit_residuals import \
    polfit_residuals_with_sigma_rejection


class CCDLine:
    """Generic CCD line definition."""

    def __init__(self):
        # set availability to False
        self.available = False
        # bounding box
        self.bb_nc1_orig = None
        self.bb_nc2_orig = None
        self.bb_ns1_orig = None
        self.bb_ns2_orig = None
        # extreme points
        self.xlower_line = None
        self.ylower_line = None
        self.xupper_line = None
        self.yupper_line = None
        # polynomial fit
        self.poly_funct = None

    def __str__(self):
        """Define printable representation of a CCDLine instance."""

        output = '<' + str(self.__class__) + " instance>"
        for key in sorted(self.__dict__.keys()):
            output += '\n=> ' + key + ': ' + str(self.__dict__[key])

        return output

    def fit(self, x, y, deg, w=None, y_vs_x=True,
            times_sigma_reject=None, title=None, debugplot=0):
        """Update the arc line from least squares fit to data.

        Parameters
        ----------
        x : 1d numpy array, float
            X coordinates of the data being fitted.
        y : 1d numpy array, float
            Y coordinates of the data being fitted.
        deg : int
            Degree of the fitting polynomial.
        w : 1d numpy array, float
            Weights to be employed in the polynomial fit.
        y_vs_x : bool
            If True, the fit is Y vs X. Otherwise, X vs Y is computed.
        times_sigma_reject : float
            If not None, deviant point are rejected.
        title : string
            Plot title.
        debugplot : int
            Determines whether intermediate computations and/or plots
            are displayed:
            00 : no debug, no plots
            01 : no debug, plots without pauses
            02 : no debug, plots with pauses
            10 : debug, no plots
            11 : debug, plots without pauses
            12 : debug, plots with pauses

        """

        # protections
        if type(x) is not np.ndarray:
            raise ValueError("x=" + str(x) + " must be a numpy.ndarray")
        if type(y) is not np.ndarray:
            raise ValueError("y=" + str(y) + " must be a numpy.ndarray")
        if x.size != y.size:
            raise ValueError("x.size != y.size")
        if w is not None:
            if type(w) is not np.ndarray:
                raise ValueError(
                    "w=" + str(w) + " must be None or a numpy.ndarray")
            if w.size != x.size:
                raise ValueError("w.size != x.size")
        if not isinstance(deg, numbers.Integral):
            raise ValueError("deg=" + str(deg) +
                             " is not a valid integer")

        # update bounding box of the CCD line
        self.bb_nc1_orig = min(x)
        self.bb_nc2_orig = max(x)
        self.bb_ns1_orig = min(y)
        self.bb_ns2_orig = max(y)

        # compute polynomial from fit to data
        if y_vs_x:
            if times_sigma_reject is None:
                # fit using the minimal domain that covers the x data
                poly_funct = \
                    Polynomial.fit(x=x, y=y, deg=deg, w=w,
                                   domain=None, window=None, full=False)
                # restore the class domain
                self.poly_funct = Polynomial.cast(poly_funct)
                # display resulting fit when requested
                if debugplot % 10 != 0:
                    polfit_residuals(x=x, y=y, deg=deg,
                                     title=title,
                                     debugplot=debugplot)
            else:
                self.poly_funct, yres_dum, reject_dum = \
                    polfit_residuals_with_sigma_rejection(
                        x=x, y=y, deg=deg,
                        title=title,
                        times_sigma_reject=times_sigma_reject,
                        debugplot=debugplot
                    )
            self.xlower_line = self.bb_nc1_orig
            self.ylower_line = self.poly_funct(self.xlower_line)
            self.xupper_line = self.bb_nc2_orig
            self.yupper_line = self.poly_funct(self.xupper_line)
        else:
            if times_sigma_reject is None:
                # fit using the minimal domain that covers the y data
                poly_funct = \
                    Polynomial.fit(x=y, y=x, deg=deg, w=w,
                                   domain=None, window=None, full=False)
                # restore the class domain
                self.poly_funct = Polynomial.cast(poly_funct)
                # display resulting fit when requested
                if debugplot % 10 != 0:
                    polfit_residuals(x=y, y=x, deg=deg,
                                     title=title, debugplot=debugplot)
            else:
                self.poly_funct, yres_dum, reject_dum = \
                    polfit_residuals_with_sigma_rejection(
                        x=y, y=x, deg=deg,
                        title=title,
                        times_sigma_reject=times_sigma_reject,
                        debugplot=debugplot
                    )
            self.ylower_line = self.bb_ns1_orig
            self.xlower_line = self.poly_funct(self.ylower_line)
            self.yupper_line = self.bb_ns2_orig
            self.xupper_line = self.poly_funct(self.yupper_line)

        # CCD line has been defined
        self.available = True

    def linspace_pix(self, start=None, stop=None, pixel_step=1, y_vs_x=None):
        """Return x,y values evaluated with a given pixel step.

        The returned values are computed within the corresponding
        bounding box of the line.

        Parameters
        ----------
        start : float
            Minimum pixel coordinate to evaluate the independent
            variable.
        stop : float
            Maximum pixel coordinate to evaluate the independent
            variable.
        pixel_step : float
            Pixel step employed to evaluate the independent variable.
        y_vs_x : bool
            If True, the polynomial fit is assumed to be Y vs X.
            Otherwise, X vs Y is employed.

        Returns
        -------
        x : 1d numpy array
            X coordinates.
        y : 1d numpy array
            Y coordinates.
        """

        if y_vs_x:
            if start is None:
                xmin = self.bb_nc1_orig
            else:
                xmin = start
            if stop is None:
                xmax = self.bb_nc2_orig
            else:
                xmax = stop
            num = int(float(xmax-xmin+1)/float(pixel_step)+0.5)
            x = np.linspace(start=xmin, stop=xmax, num=num)
            y = self.poly_funct(x)
        else:
            if start is None:
                ymin = self.bb_ns1_orig
            else:
                ymin = start
            if stop is None:
                ymax = self.bb_ns2_orig
            else:
                ymax = stop
            num = int(float(ymax-ymin+1)/float(pixel_step)+0.5)
            y = np.linspace(start=ymin, stop=ymax, num=num)
            x = self.poly_funct(y)

        return x, y

    def offset(self, offset_value):
        """Return a copy of self, shifted a constant offset.

        Parameters
        ----------
        offset_value : float
            Number of pixels to shift the CCDLine.

        """

        new_instance = deepcopy(self)
        new_instance.poly_funct.coef[0] += offset_value
        return new_instance


class ArcLine(CCDLine):
    """Arc line definition."""

    def __init__(self):
        # members of superclass CCDLine
        CCDLine.__init__(self)
        # rectified abscissa of arc line
        self.x_rectified = None

    def linspace_pix(self, start=None, stop=None, pixel_step=1, y_vs_x=False):
        """Return x,y values evaluated with a given pixel step."""
        return CCDLine.linspace_pix(self, start=start, stop=stop,
                                    pixel_step=pixel_step, y_vs_x=y_vs_x)


class SpectrumTrail(CCDLine):
    """Spectrum trail definition"""

    def __init__(self):
        # members of superclass CCDLine
        CCDLine.__init__(self)
        # rectified ordinate of spectrum trail
        self.y_rectified = None

    def linspace_pix(self, start=None, stop=None, pixel_step=1, y_vs_x=True):
        """Return x,y values evaluated with a given pixel step."""
        return CCDLine.linspace_pix(self, start=start, stop=stop,
                                    pixel_step=pixel_step, y_vs_x=y_vs_x)


def intersection_spectrail_arcline(spectrail, arcline):
    """Compute intersection of spectrum trail with arc line.

    Parameters
    ----------
    spectrail : SpectrumTrail object
        Instance of SpectrumTrail class.
    arcline : ArcLine object
        Instance of ArcLine class

    Returns
    -------
    xroot, yroot : tuple of floats
        (X,Y) coordinates of the intersection.

    """

    # approximate location of the solution
    expected_x = (arcline.xlower_line + arcline.xupper_line) / 2.0

    # composition of polynomials to find intersection as
    # one of the roots of a new polynomial
    rootfunct = arcline.poly_funct(spectrail.poly_funct)
    rootfunct.coef[1] -= 1
    # compute roots to find solution
    tmp_xroots = rootfunct.roots()

    # take the nearest root to the expected location
    xroot = tmp_xroots[np.abs(tmp_xroots - expected_x).argmin()]
    if np.isreal(xroot):
        xroot = xroot.real
    else:
        raise ValueError("xroot=" + str(xroot) +
                         " is a complex number")
    yroot = spectrail.poly_funct(xroot)

    return xroot, yroot



#
# Copyright 2015-2023 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#

import sys
import os
import pytest

from .testcache import user_cache_dir


@pytest.mark.skipif(sys.platform != 'linux2',
                    reason="runs only in linux")
def test_user_cache_dir_linux_home(monkeypatch, tmpdir):

    home = tmpdir.mkdir('hometest')

    modenviron = {'HOME': home.strpath}

    monkeypatch.setattr(os, 'environ', modenviron)

    cachedir = os.path.join(home.strpath, '.cache')
    expected = os.path.join(cachedir, 'numina')

    if not os.path.exists(expected):
        os.makedirs(expected)

    assert user_cache_dir("numina") == expected
    assert os.path.exists(os.path.join(expected, 'astropy'))


@pytest.mark.skipif(sys.platform != 'linux2',
                    reason="runs only in linux")
def test_user_cache_dir_linux_xdg(monkeypatch, tmpdir):

    home = tmpdir.mkdir('hometest')
    cache = tmpdir.mkdir('.cache')

    modenviron = {'HOME': home.strpath,
                  'XDG_CACHE_HOME': cache.strpath
                  }

    monkeypatch.setattr(os, 'environ', modenviron)

    cachedir = cache.strpath
    expected = os.path.join(cachedir, 'numina')

    assert user_cache_dir("numina") == expected
    assert os.path.exists(os.path.join(expected, 'astropy'))


@pytest.mark.skipif(sys.platform != 'darwin',
                    reason="runs only in darwin")
def test_user_cache_dir_darwin_home(monkeypatch, tmpdir):

    home = tmpdir.mkdir('hometest')

    modenviron = {'HOME': home.strpath}

    monkeypatch.setattr(os, 'environ', modenviron)

    cachedir = os.path.join(home.strpath, 'Library/Caches')
    expected = os.path.join(cachedir, 'numina')

    if not os.path.exists(expected):
        os.makedirs(expected)

    assert user_cache_dir("numina") == expected
    assert os.path.exists(os.path.join(expected, 'astropy'))


@pytest.mark.skipif(sys.platform != 'darwin',
                    reason="runs only in darwin")
def test_user_cache_dir_darwin_xdg(monkeypatch, tmpdir):

    home = tmpdir.mkdir('hometest')
    cache = tmpdir.mkdir('.cache')

    modenviron = {'HOME': home.strpath,
                  'XDG_CACHE_HOME': cache.strpath
                  }

    monkeypatch.setattr(os, 'environ', modenviron)

    cachedir = os.path.join(home.strpath, 'Library/Caches')
    expected = os.path.join(cachedir, 'numina')

    assert user_cache_dir("numina") == expected
    assert os.path.exists(os.path.join(expected, 'astropy'))



#
# Copyright 2016-2020 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#

"""A monotonic piecewise cubic interpolator."""

import numpy as np


class SteffenInterpolator(object):
    """
    A monotonic piecewise cubic 1-d interpolator.

    A monotonic piecewise cubic interpolator based on
    Steffen, M., Astronomy & Astrophysics, 239, 443-450 (1990)

    `x` and `y` are arrays of values used to approximate some function f: `y = f(x)`.
    This class returns an object whose call method uses
    monotonic cubic splines to find the value of new points.

    Parameters
    ----------
    x : (N,), array_like
        A 1-D array of real values, sorted monotonically increasing.
    y : (N,), array_like
        A 1-D array of real values.
    yp_0 : float, optional
        The value of the derivative in the first sample.
    yp_N : float, optional
        The value of the derivative in the last sample.
    extrapolate : str, optional
        Specifies the kind of extrapolation as a string
        ('extrapolate', 'zeros', 'raise', 'const', 'border')
        If 'raise' is set, when interpolated values are requested outside of the
        domain of the input data (x,y), a ValueError is raised.
        If 'const' is set, 'fill_value' is returned.
        If 'zeros' is set, '0' is returned.
        If 'border' is set, 'y[0]' is returned for values below 'x[0]'
        and 'y[N-1]' is returned for values above 'x[N-1]'
        If 'extrapolate' is set, the extreme polynomial are extrapolated
        outside of their ranges.
        Default is 'raise'.
    fill_value : float, optional
        If provided, then this value will be used to fill in for requested
        points outside of the data range when `extrapolation`is set to `const`.
        If not provided, then the default is NaN.

    """

    def __init__(self, x, y, yp_0=0.0, yp_N=0.0, extrapolate='raise', fill_value=np.nan):

        # 'zeros' is a shortcut
        if extrapolate == 'zeros':
            extrapolate = 'const'
            fill_value = 0.0

        self.fill_value = fill_value
        self._extrapolation(extrapolate)

        # Compute all coefficients

        N = len(x)

        # Steps
        h = self._create_h(x)
        # Secants
        s = self._create_s(y, h)

        abs_s = np.abs(s)

        # Check monotonicity given borders
        if yp_0 * s[0] < 0:
            raise ValueError('monotonicity not guaranteed')

        if abs(yp_0) > 3*abs_s[0]:
            raise ValueError('monotonicity not guaranteed')

        if yp_N * s[-2] < 0:
            raise ValueError('monotonicity not guaranteed')

        if abs(yp_N) > 3*abs_s[-2]:
            raise ValueError('monotonicity not guaranteed')

        p = self._create_p(s, h)

        # Derivatives
        yp = np.zeros((N+1,))
        yp[0] = yp_0
        yp[N] = yp_N
        sign_s = np.sign(s)
        min_abs_s = np.minimum(abs_s[1:], abs_s[:-1])
        yp[1:-1] = (sign_s[1:] + sign_s[:-1]) * \
            np.minimum(min_abs_s, 0.5 * np.abs(p[1:]))

        # Polynomial coefficients
        self._d = y
        self._c = yp[:-1]
        self._b = (3 * s - 2 * yp[:-1] - yp[1:]) / h
        self._a = (yp[:-1] + yp[1:] - 2 * s) / h**2
        # Other values
        self._x = x

    def _extrapolation(self, extrapolate):
        """Check permitted values of extrapolation."""
        modes = ['extrapolate',
                 'raise',
                 'const',
                 'border']
        if extrapolate not in modes:
            msg = f'invalid extrapolation mode {extrapolate}'
            raise ValueError(msg)

        if extrapolate == 'raise':
            self.bounds_error = True
            self.extrapolate = False
        else:
            self.extrapolate = True
            self.bounds_error = False
        self.extrapolate_mode = extrapolate

    @staticmethod
    def _create_s(y, h):
        """Estimate secants"""
        s = np.zeros_like(y)
        s[:-1] = (y[1:] - y[:-1]) / h[:-1]
        s[-1] = 0.0
        return s

    @staticmethod
    def _create_h(x):
        """increase between samples"""
        h = np.zeros_like(x)
        h[:-1] = x[1:] - x[:-1]
        # border
        h[-1] = h[-2]
        return h

    @staticmethod
    def _create_p(s, h):
        """Parabolic derivative"""
        p = np.zeros_like(s)
        p[1:] = (s[:-1]*h[1:] + s[1:] * h[:-1]) / (h[1:] + h[:-1])
        return p

    def _eval(self, v, in_bounds, der):
        """Eval polynomial inside bounds."""
        result = np.zeros_like(v, dtype='float')
        x_indices = np.searchsorted(self._x, v, side='right')
        ids = x_indices[in_bounds] - 1
        u = v[in_bounds] - self._x[ids]
        result[in_bounds] = self._poly_eval(u, ids, der)
        return result

    def _extrapolate(self, result, v, below_bounds, above_bounds, der):
        """Extrapolate result based on extrapolation mode."""
        if self.extrapolate_mode == 'const':
            fill_b = fill_a = self.fill_value
        elif self.extrapolate_mode == 'border':
            fill_b = self._poly_eval(0, 0, der)
            fill_a = self._poly_eval(0, -1, der)
        elif self.extrapolate_mode == 'extrapolate':
            u = v[above_bounds] - self._x[-2]
            fill_a = self._poly_eval(u, -2, der)
            u = v[below_bounds] - self._x[0]
            fill_b = self._poly_eval(u, 0, der)
        else:
            raise ValueError("extrapolation method doesn't exist")

        result[below_bounds] = fill_b
        result[above_bounds] = fill_a

    def __call__(self, x_new, der=0):
        """
        Evaluate the Steffen interpolant

        Parameters
        ----------
        x_new : array-like
              Points to evaluate the interpolant at.

        der : int, optional
              Degree of the derivative. 0 is the plain interpolant.
              For der >= 4, the return value is 0
        Returns
        -------
        y : array-like
            Interpolated values.

        """

        v = np.asarray(x_new)

        below_bounds, above_bounds = self._check_bounds(v)
        in_bounds = np.logical_and(~below_bounds, ~above_bounds)

        result = self._eval(v, in_bounds, der)

        if self.extrapolate:
            self._extrapolate(result, v, below_bounds, above_bounds, der)
        return result

    def _poly_eval(self, u, ids, der=0):
        """Evaluate internal polynomial."""
        if der == 0:
            return self._poly_eval_0(u, ids)
        elif der == 1:
            return self._poly_eval_1(u, ids)
        elif der == 2:
            return self._poly_eval_2(u, ids)
        elif der == 3:
            return self._poly_eval_3(u, ids)
        elif der >= 4:
            return self._poly_eval_4(u, ids)
        else:
            raise ValueError(f"der={der} is impossible")

    def _poly_eval_0(self, u, ids):
        """Evaluate internal polynomial."""
        return u * (u * (self._a[ids] * u + self._b[ids]) + self._c[ids]) + self._d[ids]

    def _poly_eval_1(self, u, ids):
        """Evaluate internal polynomial."""
        return u * (3*self._a[ids] * u + 2*self._b[ids]) + self._c[ids]

    def _poly_eval_2(self, u, ids):
        """Evaluate internal polynomial."""
        return 6 * self._a[ids] * u + 2*self._b[ids]

    def _poly_eval_3(self, u, ids):
        """Evaluate internal polynomial."""
        return 6 * self._a[ids]

    def _poly_eval_4(self, u, ids):
        """Evaluate internal polynomial."""
        return u * 0.0

    def _check_bounds(self, v):
        """Check which values are out of bounds.

        Raises
        ------
        ValueError:

        """
        below_bounds = v < self._x[0]
        above_bounds = v > self._x[-1]

        if self.bounds_error and below_bounds.any():
            raise ValueError("A value in x_new is below the interpolation "
                             "range.")
        if self.bounds_error and above_bounds.any():
            raise ValueError("A value in x_new is above the interpolation "
                             "range.")

        return below_bounds, above_bounds

# A possible optimization if the points are in a regular grid
# Is a bit faster (~2 times)

# class SteffenInterpolatorGrid(SteffenInterpolator):
#     @staticmethod
#     def _create_h(x):
#         return x[1] - x[0]
#
#     @staticmethod
#     def _create_p(s, h):
#         p = np.zeros_like(s)
#         p[1:] = (s[:-1] + s[1:]) / 2.0
#         return p
#
#     @staticmethod
#     def _create_s(y, h):
#         s = np.zeros_like(y)
#         s[:-1] = (y[1:] - y[:-1]) / h
#         s[-1] = 0.0
#         return s
#
#     def __init__(self, x, y, yp_0=0.0, yp_N=0.0, extrapolate='raise',
#                  fill_value=np.nan):
#         super(SteffenInterpolatorGrid, self).__init__(x, y, yp_0, yp_N,
#                                                       extrapolate=extrapolate,
#                                                       fill_value=fill_value)
#
#         self.h = self._create_h(x)
#
#     def _eval(self, v, in_bounds):
#         result = np.zeros_like(v, dtype='float')
#         off = (v - self._x[0]) / self.h
#         x_indices1 = np.floor(off).astype('int')
#         ids = x_indices1[in_bounds]
#         u = v[in_bounds] - (self._x[0] + ids * self.h)
#         result[in_bounds] = self._poly_eval(u, ids)
#         return result
#



#
# Copyright 2016-2023 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#

"""DAL for file-based database of products."""

import os
import itertools
import logging

# import numina.util.objimport as objimp
from numina.store import load
from numina.exceptions import NoResultFound
from numina.dal import StoredProduct

_logger = logging.getLogger(__name__)


def _combinations(seq):
    maxl = len(seq)
    for i in range(maxl + 1, -1, -1):
        for j in itertools.combinations(seq, i):
            yield j


class FileFinder(object):

    def candidates(self, directory):
        return [filename for filename in sorted(os.listdir(directory))]

    def build_path(self, directory, value):
        return os.path.join(directory, value)

    def build_final_path(self, directory, value):
        return os.path.join(directory, value)

    def check(self, directory, value):
        fname = value
        loadpath = self.build_path(directory, fname)
        # print('loadpath=', loadpath)
        _logger.debug("check %s", loadpath)
        if fname.startswith("."):
            _logger.debug("file %s is hidden, ignore", loadpath)
            return False
        if os.path.isfile(loadpath):
            _logger.debug("is regular file %s", loadpath)
            _logger.info("found %s", loadpath)
            # print(loadpath)
            return True
        else:
            _logger.debug("is not regular file %s", loadpath)
            return False


class FileFinderGTC(FileFinder):

    def candidates(self, directory):
        base = [('result.json', 1)]
        other = [(m, 0)
                 for m in super(FileFinderGTC, self).candidates(directory)]
        base.extend(other)
        return base

    def build_path(self, directory, value):
        return super(FileFinderGTC, self).build_path(directory, value)

    def build_final_path(self, directory, value):
        return os.path.join(directory, value[0]), value[1]

    def check(self, directory, value):
        fname, kind = value
        return super(FileFinderGTC, self).check(directory, fname)


def build_product_path(drp, rootdir, conf, name, tipo, ob, cls=FileFinderGTC):

    _logger.info('search %s of type %s', name, tipo)

    file_finder = cls()

    try:
        # FIXME
        res = drp.query_provides(tipo.__class__)
        label = res.alias
    except ValueError:
        label = tipo.name()

    # search results of these OBs
    # build path based in combinations of tags
    vals = [ob.tags[k] for k in sorted(ob.tags.keys())]
    for com in _combinations(vals):
        directory = os.path.join(rootdir, ob.instrument, conf, label, *com)
        _logger.debug('try directory %s', directory)
        try:
            for value in file_finder.candidates(directory):
                if file_finder.check(directory, value):
                    return file_finder.build_final_path(directory, value)
        except OSError as msg:
            _logger.debug("%s", msg)
    else:
        msg = f'type {label} compatible with tags {ob.tags!r} not found'
        _logger.info("%s", msg)
        raise NoResultFound(msg)


DAL_USE_OFFLINE_CALIBS = True


class DiskFileDAL(object):

    def __init__(self, drp, rootdir, *args, **kwargs):
        super(DiskFileDAL, self).__init__()
        self.drp = drp
        self.rootdir = rootdir
        self.conf = 'default'

    def search_product(self, name, tipo, ob):
        klass = tipo.__class__
        print('Init search ', name, tipo, tipo.__class__)

        try:
            res = self.drp.query_provides(tipo.__class__)
            label = res.alias
        except ValueError:
            label = tipo.__class__.__name__

        # search results of these OBs
        # build path based in combinations of tags
        for com in _combinations(ob.tags.values()):
            directory = os.path.join(
                self.rootdir, ob.instrument, self.conf, label, *com)
            print('search in', directory)
            try:
                files_s = [filename for filename in sorted(
                    os.listdir(directory))]
                print("files_s", files_s)
                for fname in files_s:
                    loadpath = os.path.join(directory, fname)
                    print("check ", loadpath)
                    if os.path.isfile(loadpath):
                        print("is regular file ", loadpath)
                        if DAL_USE_OFFLINE_CALIBS:
                            content = load(tipo, loadpath)
                        # else:
                            # data = json.load(open(loadpath))
                            # result = process_result(data)
                            # key = self._field_to_extract[klass]
                            # content = result[key]
                            return StoredProduct(id=files_s[-1], content=content, tags=ob.tags)
                        else:
                            print("not ready")
                    else:
                        print("is not regular file ", loadpath)
            except OSError as msg:
                print(msg)
                # msg = 'type %s compatible with tags %r not found' % (klass, ob.tags)
                # raise NoResultFound(msg)
        else:
            msg = f'type {klass} compatible with tags {ob.tags!r} not found'
            print(msg)
            raise NoResultFound(msg)




import numpy
import numpy.random

from ..background import background_estimator
from ..background import create_background_map


def test_back1():
    bdata = 300 * numpy.ones((50, 50))
    back, std = background_estimator(bdata)
    assert numpy.allclose([back, std], [300, 0])


def test_back2():
    res = (301.23398492452793, 12.277994974842644)
    bdata = 300 * numpy.ones((50, 50))
    bdata[20:25, 20:25] = 700
    r = background_estimator(bdata)
    assert numpy.allclose(r, res)


def test_back_crowded():
    res = (318.5390526053346, 73.37968420177819)
    bdata = 300 * numpy.ones((50, 50))
    bdata[20:25, 20:25] = 700
    bdata[30:35, 30:35] = 700
    bdata[10:15, 10:15] = 700
    bdata[10:15, 20:25] = 700
    bdata[10:15, 30:35] = 700
    bdata[20:25, 30:35] = 700
    r = background_estimator(bdata)
    assert numpy.allclose(r, res)


def test_background_map():
    numpy.random.seed(seed=938483)
    bck = numpy.random.normal(1100, 32, (512, 512))
    nd, ns = create_background_map(bck, 8, 8)
    res = (1100.055068749893, 31.478283226675636)
    assert numpy.allclose([nd.mean(), ns.mean()], res)



def test_configure1(pattri_ins, pattri_header):
    pattri_ins.configure_with_header(pattri_header)
    assert pattri_header["INSMODE"] == "Mode_B"
    assert pattri_ins.get_property("PSU.mode") == "Mode_B"


def test_configure2(pattri_ins, pattri_header):
    assert pattri_header["INSMODE"] == "Mode_B"
    pattri_ins.configure_with_header(pattri_header)
    assert pattri_ins.get_property("PSU.spaces") == []


def test_configure1b(pattri_ins, pattri_state):
    pattri_ins.configure(pattri_state)
    assert pattri_ins.get_property("PSU.mode") == "Mode_B"
    assert pattri_ins.get_property("PSU.spaces") == []



import pytest
from numina.tests.seffect import record_call

from ..state import State, Status, Transition


def test_state1():
    state = State()
    assert state.current == Status.ACTIVE


def test_configuring():
    state = State()
    state.transition(Transition.CONFIGURE)
    assert state.current == Status.CONFIGURING
    state.transition(Transition.END_CONFIGURE)
    assert state.current == Status.ACTIVE


def test_configuring2():
    state = State()
    with state.managed_configure():
        try:
            raise ValueError()
        except ValueError:
            state.transition(Transition.ERROR)

    assert state.current == Status.FAILED


def test_configuring3():
    state = State()
    with state.managed_configure():
        assert state.current == Status.CONFIGURING

    assert state.current == Status.ACTIVE


def test_reset1():

    @record_call
    def enter_failed_cb():
        pass

    @record_call
    def exit_failed_cb():
        pass

    @record_call
    def enter_reset_cb():
        pass

    state = State()
    state.enter_failed.connect(enter_failed_cb)
    state.exit_failed.connect(exit_failed_cb)
    state.enter_reset.connect(enter_reset_cb)
    state.transition(Transition.ERROR)

    assert state.current == Status.FAILED
    assert enter_failed_cb.side_effect.called is True
    assert exit_failed_cb.side_effect.called is False
    assert enter_reset_cb.side_effect.called is False

    for t_name in [
        Transition.CONFIGURE,
        Transition.END_CONFIGURE,
        Transition.END_RESET,
    ]:
        enter_failed_cb.side_effect.clear()
        exit_failed_cb.side_effect.clear()
        enter_reset_cb.side_effect.clear()
        state.transition(t_name)
        assert state.current == Status.FAILED
        assert enter_failed_cb.side_effect.called is False
        assert exit_failed_cb.side_effect.called is False
        assert enter_reset_cb.side_effect.called is False

    enter_failed_cb.side_effect.clear()
    exit_failed_cb.side_effect.clear()
    enter_reset_cb.side_effect.clear()
    state.transition(Transition.RESET)
    assert state.current == Status.RESETTING
    assert enter_failed_cb.side_effect.called is False
    assert exit_failed_cb.side_effect.called is True
    assert enter_reset_cb.side_effect.called is True


def create_state():
    created_signals = {}
    for n in [
        "enter_failed",
        "exit_failed",
        "enter_reset",
        "exit_reset",
        "enter_active",
        "enter_configure",
    ]:
        created_signals[n] = record_call(lambda: None)

    state = State()
    for key in created_signals:
        getattr(state, key).connect(created_signals[key])

    return state, created_signals


@pytest.mark.parametrize(
    "t_name",
    [
        Transition.CONFIGURE,
        Transition.END_CONFIGURE,
    ],
)
def test_reset_fail(t_name):

    state, created_signals = create_state()

    state.transition(Transition.RESET)

    for n in created_signals:
        created_signals[n].side_effect.clear()

    state.transition(t_name)
    assert state.current == Status.FAILED

    called_signals = {}
    for n in created_signals:
        called_signals[n] = False
    called_signals["enter_failed"] = True
    called_signals["exit_reset"] = True
    for n in called_signals:
        assert created_signals[n].side_effect.called == called_signals[n]


def test_reset_active():

    state, created_signals = create_state()

    state.transition(Transition.RESET)

    for n in created_signals:
        created_signals[n].side_effect.clear()

    state.transition(Transition.END_RESET)
    assert state.current == Status.ACTIVE

    called_signals = {}
    for n in created_signals:
        called_signals[n] = False
    called_signals["enter_active"] = True
    called_signals["exit_reset"] = True
    for n in called_signals:
        assert created_signals[n].side_effect.called == called_signals[n]


def test_reset_reset():

    state, created_signals = create_state()

    state.transition(Transition.RESET)

    for n in created_signals:
        created_signals[n].side_effect.clear()

    state.transition(Transition.RESET)
    assert state.current == Status.RESETTING

    called_signals = {}
    for n in created_signals:
        called_signals[n] = False
    for n in called_signals:
        assert created_signals[n].side_effect.called == called_signals[n]


def test_active_reset0():
    """From active to reset"""
    state, created_signals = create_state()

    state.transition(Transition.RESET)

    called_signals = {}
    for n in created_signals:
        called_signals[n] = False
    called_signals["enter_reset"] = True

    assert state.current == Status.RESETTING
    for n in called_signals:
        assert created_signals[n].side_effect.called == called_signals[n]


def test_active_config0():
    """From active to configure"""
    state, created_signals = create_state()

    state.transition(Transition.CONFIGURE)

    called_signals = {}
    for n in created_signals:
        called_signals[n] = False
    called_signals["enter_configure"] = True

    assert state.current == Status.CONFIGURING
    for n in called_signals:
        assert created_signals[n].side_effect.called == called_signals[n]


def test_config_config():

    state, created_signals = create_state()

    state.transition(Transition.CONFIGURE)

    for n in created_signals:
        created_signals[n].side_effect.clear()

    state.transition(Transition.CONFIGURE)
    assert state.current == Status.CONFIGURING

    called_signals = {}
    for n in created_signals:
        called_signals[n] = False
    for n in called_signals:
        assert created_signals[n].side_effect.called == called_signals[n]



#
# Copyright 2015-2023 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#

"""Store the solution of a wavelength calibration"""

import math
import warnings

from numpy.polynomial import Polynomial


class CrLinear(object):
    """Store information concerning the linear wavelength calibration.

    Parameters
    ----------
    crpix : float
        CRPIX1 value employed in the linear wavelength calibration.
    crval : float
        CRVAL1 value corresponding tot he linear wavelength
        calibration.
    crmin : float
        CRVAL value at pixel number 1 corresponding to the linear
        wavelength calibration.
    crmax : float
        CRVAL value at pixel number NAXIS1 corresponding to the linear
        wavelength calibration.
    cdelt : float
        CDELT1 value corresponding to the linear wavelength
        calibration.

    Attributes
    ----------
    Identical to parameters.

    """

    def __init__(self, crpix, crval, crmin, crmax, cdelt):
        self.crpix = crpix
        self.crval = crval
        self.cdelt = cdelt
        self.crmin = crmin
        self.crmax = crmax

    def __getstate__(self):
        state = self.__dict__.copy()
        for k in state:
            state[k] = float(state[k])
        return state

    def __str__(self):
        """Printable representation of a CrLinear instance."""

        output = "<CrLinear instance>\n" + \
                 "crpix linear: " + str(self.crpix) + "\n" + \
                 "crval linear: " + str(self.crval) + "\n" + \
                 "cdelt linear: " + str(self.cdelt) + "\n" + \
                 "crmin linear: " + str(self.crmin) + "\n" + \
                 "crmax linear: " + str(self.crmax)

        return output


class WavecalFeature(object):
    """Store information concerning a particular line identification.

    Parameters
    ----------
    line_ok : bool
        True if the line has been properly identified.
    category : char
        Line identification type (A, B, C, D, E, R, T, P, K, I, X).
        See documentation embedded within the arccalibration_direct
        function for details.
    lineid : int
        Number of identified line within the master list.
    xpos : float
        Pixel x-coordinate of the peak of the line.
    ypos : float
        Pixel y-coordinate of the peak of the line.
    peak : float
        Flux of the peak of the line.
    fwhm : float
        FWHM of the line.
    reference : float
        Wavelength of the identified line in the master list.
    wavelength : float
        Wavelength of the identified line estimated from the wavelength
        calibration polynomial.
    funcost : float
        Cost function corresponding to each identified arc line.

    Attributes
    ----------
    Identical to parameters.

    """

    def __init__(self, line_ok, category, lineid, funcost, xpos, ypos=0.0,
                 peak=0.0, fwhm=0.0, reference=0.0, wavelength=0.0):
        self.line_ok = line_ok
        self.category = category
        self.lineid = lineid
        self.funcost = funcost
        self.xpos = xpos
        self.ypos = ypos
        self.peak = peak
        self.fwhm = fwhm
        self.reference = reference
        self.wavelength = wavelength

    def __getstate__(self):
        state = self.__dict__.copy()
        float_keys = ['funcost', 'xpos', 'ypos', 'peak', 'fwhm',
                      'reference', 'wavelength']
        for k in state:
            if k in float_keys:
                value = float(state[k])
                # translate infinities
                if math.isinf(value):
                    value = 1e50
                    warnings.warn(
                        f'Converting {k}=inf to {value}',
                        RuntimeWarning
                    )

                state[k] = value
            elif k in ['lineid']:
                state[k] = int(state[k])
            else:
                pass
        return state

    def __str__(self):
        if self.line_ok:
            sline_ok = 'True '
        else:
            sline_ok = 'False'
        output = "<WavecalFeature instance>\n" + \
                 " line_ok: " + sline_ok + "  " + \
                 "category: " + str(self.category) + "  " + \
                 "id: " + str(self.lineid) + "  " + \
                 "xpos: " + str(self.xpos) + "  " + \
                 "ypos: " + str(self.ypos) + "\n" + \
                 " peak: " + str(self.peak) + "  " + \
                 "fwhm: " + str(self.fwhm) + "  " + \
                 "reference: " + str(self.reference) + "  " + \
                 "wavelength: " + str(self.wavelength) + "  " + \
                 "funcost: " + str(self.funcost)

        return output


class SolutionArcCalibration(object):
    """Auxiliary class to store the arc calibration solution.

    Note that this class only stores the information concerning the
    arc lines that have been properly identified. The information
    about all the lines (including those initially found but at the
    end discarded) is stored in the list of WavecalFeature instances.

    Parameters
    ----------
    features : list (of WavecalFeature instances)
        A list of size equal to the number of identified lines, which
        elements are instances of the class WavecalFeature, containing
        all the relevant information concerning the line
        identification.
    coeff : 1d numpy array (float)
        Coefficients of the wavelength calibration polynomial.
    residual_std : float
        Residual standard deviation of the fit.
    cr_linear : instance of CrLinear
        Object containing the linear approximation parameters crpix,
        crval, cdelt, crmin and crmax.

    Attributes
    ----------
    Identical to parameters.

    """

    def __init__(self, features, coeff, residual_std, cr_linear):

        self.features = features
        self.coeff = [float(tmpcoeff) for tmpcoeff in coeff]
        # force residual_std to be a float and not a scalar numpy array
        self.residual_std = float(residual_std)
        self.cr_linear = cr_linear

        # use fitted polynomial to predict wavelengths at the line
        # locations given by xpos
        self.update_features(poly=Polynomial(self.coeff))

    def update_features(self, poly):
        """Evaluate wavelength at xpos using the provided polynomial."""

        for feature in self.features:
            feature.wavelength = poly(feature.xpos)

    @property
    def nlines_arc(self):
        return len([wvfeature for wvfeature in self.features
                    if wvfeature.line_ok])

    def __eq__(self, other):
        if isinstance(other, SolutionArcCalibration):
            return self.__getstate__() == other.__getstate__()
        else:
            return NotImplemented

    def __ne__(self, other):
        return not self == other

    def __str__(self):
        """Printable representation of a SolutionArcCalibration instance."""

        output = "<SolutionArcCalibration instance>\n" + \
                 "- Number arc lines: " + str(self.nlines_arc) + "\n" + \
                 "- Coeff...........: " + str(self.coeff) + "\n" + \
                 "- Residual std....: " + str(self.residual_std) + "\n" + \
                 "- " + str(self.cr_linear) + "\n"

        for feature in self.features:
            output += "- " + str(feature) + "\n"

        # return string with all the information
        return output

    def __getstate__(self):
        result = self.__dict__.copy()
        result['features'] = [
            feature.__getstate__() for feature in self.features
        ]
        result['cr_linear'] = result['cr_linear'].__getstate__()
        return result

    def __setstate__(self, state):
        self.__dict__ = state.copy()
        features = []
        for feature in state['features']:
            newf = WavecalFeature.__new__(WavecalFeature)
            newf.__dict__ = feature
            features.append(newf)

        self.features = features
        self.cr_linear = CrLinear(**state['cr_linear'])



import pytest

import numina.exceptions
import numina.core.recipes
from ..validator import validate
from ..validator import only_positive
from ..validator import as_list
from ..validator import range_validator


class RecipeIO(object):
    def __init__(self, valid=True):
        self.valid = valid
        self.was_called = False

    def validate(self):
        self.was_called = True
        if not self.valid:
            raise numina.exceptions.ValidationError("This value is invalid")
        return True


class RecipeInput(RecipeIO):
    pass


class RecipeResult(RecipeIO):
    pass


class Recipe1(numina.core.recipes.BaseRecipe):
    @validate
    def run(self, rinput):
        return RecipeResult()


class Recipe2(numina.core.recipes.BaseRecipe):
    @validate
    def run(self, rinput):
        return RecipeResult(valid=False)


def test_validate():

    recipe = Recipe1()
    rinput = RecipeInput()
    result = recipe.run(rinput)

    assert rinput.was_called
    assert result.was_called


def test_validate_exception1():

    recipe = Recipe1()
    rinput = RecipeInput(valid=False)
    with pytest.raises(numina.exceptions.ValidationError):
        recipe.run(rinput)


def test_validate_exception2():

    recipe = Recipe2()
    rinput = RecipeInput(valid=True)
    with pytest.raises(numina.exceptions.ValidationError):
        recipe.run(rinput)


def test_only_positive1():
    """Test only_positive pass positice values"""
    for value in [0, 1, 2, 50, 400]:
        assert value == only_positive(value)


def test_only_positive2():
    """Test only_positive raises on negative values"""
    for value in [-1, -2, -50, -400, -1.3, -2939]:
        with pytest.raises(numina.exceptions.ValidationError):
            only_positive(value)


def test_as_list1():
    """Test as_list decorator"""
    @as_list
    def some_test(value):
        return value

    values = [0, 1, 2, 50, 400]
    assert values == some_test(values)


def test_as_list2():
    """Test as_list decorator raises ValidationError"""
    @as_list
    def some_test(value):
        if value > 1:
            raise numina.exceptions.ValidationError('must be <= 1')
        return value

    values = [0, 1, 2, 50, 400]
    with pytest.raises(numina.exceptions.ValidationError):
        some_test(values)


@pytest.mark.parametrize("validator, allowed, not_allowed", [
    (range_validator(), [-1, 2.0, 34.0, 4, -5], []),
    (range_validator(minval=3), [3, 34.0, 4], [2.99, -1]),
    (range_validator(maxval=3), [3, -1, -100], [3.1, 34.0, 4]),
    (range_validator(minval=2, maxval=3), [2, 2.5, 3], [-4, -1, 100])
])
def test_range_validator3(validator, allowed, not_allowed):
    """Range validators"""

    for val in allowed:
        assert val == validator(val)

    for val in not_allowed:
        with pytest.raises(numina.exceptions.ValidationError):
            validator(val)



#
# Copyright 2018 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#


import math
import numbers


class PixelInterval1D(object):
    """One dimensional bounding box with integer indices.

    Upper value (pix2) is not included in the
    box, as in slices.

    Parameters
    ----------
    pix1, pix2 : int
    axis : int

    """

    def __init__(self, pix1, pix2, axis=0):

        for name in ['pix1', 'pix2']:
            value = locals()[name]
            if not isinstance(value, numbers.Integral):
                raise TypeError(f"'{name}' must be integer in axis {axis}")

        if pix2 < pix1:
            raise ValueError(f"'pix2' must be >= 'pix1' in axis {axis}")

        self.pix1 = pix1
        self.pix2 = pix2
        self.ndim = 1
        self.axis = axis
        self._empty = (self.pix2 == self.pix1)

    @property
    def shape(self):
        return self.pix2 - self.pix1

    @property
    def slice(self):
        return slice(self.pix1, self.pix2)

    @classmethod
    def from_coordinates(cls, x1, x2):
        pix1, pix2 = cls.pixel_range(x1, x2)
        return cls(pix1, pix2)

    @classmethod
    def pixel_range(cls, x1, x2):
        iws = [int(math.floor(w + 0.5)) for w in [x1, x2]]
        pix1, pix2 = iws
        return pix1, pix2 + 1

    def union(self, other):
        if self:
            if other:
                npix2 = max(self.pix2, other.pix2)
                npix1 = min(self.pix1, other.pix1)
                return PixelInterval1D(npix1, npix2, axis=self.axis)
            else:
                return self
        else:
            return other

    def intersection(self, other):
        if self:
            if other:
                if (other.pix1 < self.pix2) and (self.pix1 < other.pix2):
                    npix1 = max(self.pix1, other.pix1)
                    npix2 = min(self.pix2, other.pix2)
                    return PixelInterval1D(npix1, npix2, axis=self.axis)
                else:
                    return PixelInterval1D(self.pix2, self.pix2, axis=self.axis)
            else:
                return self
        else:
            return other

    def __eq__(self, other):
        if isinstance(other, type(self)):
            if not (self or other):
                # if both are empty, are equal
                return True
            if self.pix1 != other.pix1:
                return False
            if self.pix2 != other.pix2:
                return False
            return True
        else:
            return NotImplemented

    def __ne__(self, other):
        return not (self == other)

    def __len__(self):
        return self.shape


class PixelInterval(object):
    """N-dimensional bounding box with integer indices.

    Upper value (pix2) is not included in the
    box, as in slices.

    Parameters
    ----------
    args : N tuples of integer pairs (pix1, pix2)

    """

    def __init__(self, *args):
        self.ndim = len(args)
        self.intervals = [PixelInterval1D(a, b, axis)
                          for axis, (a, b) in enumerate(args)]

    @property
    def shape(self):
        return tuple(intl.shape for intl in self.intervals)

    @property
    def slice(self):
        return tuple(intl.slice for intl in self.intervals)

    @classmethod
    def pixel_range(cls, *args):
        return tuple(PixelInterval1D.pixel_range(*arg)
                     for arg in args)

    @classmethod
    def from_coordinates(cls, *args):
        newargs = cls.pixel_range(*args)
        return cls(*newargs)

    @classmethod
    def from_intervals(cls, intervals):
        result = super(PixelInterval, cls).__new__(cls)
        result.intervals = list(intervals)
        # update axis
        for idx, intl in enumerate(result.intervals):
            intl.axis = idx
        result.ndim = len(result.intervals)
        return result

    def union(self, other):
        if self.ndim == other.ndim:
            intl = [s.union(o)
                    for s, o in zip(self.intervals, other.intervals)]
            return self.from_intervals(intl)
        else:
            raise NotImplementedError

    def intersection(self, other):
        if self.ndim == other.ndim:
            intl = [s.intersection(o) for s, o in zip(
                self.intervals, other.intervals)]
            return self.from_intervals(intl)
        else:
            raise NotImplementedError

    def __eq__(self, other):
        if isinstance(other, type(self)):

            if self.ndim != other.ndim:
                return False
            if not (self or other):
                # if both are empty, are equal
                return True

            return all(m == o for m, o in zip(self.intervals, other.intervals))
        else:
            return NotImplemented

    def __ne__(self, other):
        return not (self == other)

    def __len__(self):
        if self.intervals:
            result = 1
            for intl in self.intervals:
                result *= len(intl)
        else:
            result = 0
        return result


class BoundingBox(PixelInterval):
    """Rectangular bounding box with integer indices.

    Upper values (ix2 and iy2) are not included in the
    box, as in slices.

    Parameters
    ----------
    ix1, ix2, iy1, iy2 : int

    """

    def __init__(self, ix1, ix2, iy1, iy2):
        newargs = (iy1, iy2), (ix1, ix2)
        super(BoundingBox, self).__init__(*newargs)

    @classmethod
    def from_coordinates(cls, x1, x2, y1, y2):
        newargs = (y1, y2), (x1, x2)
        dum = cls.pixel_range(*newargs)
        ix1, ix2 = dum[1]
        iy1, iy2 = dum[0]
        return cls(ix1, ix2, iy1, iy2)

    @property
    def extent(self):
        """Helper for matplotlib imshow"""
        return (
            self.intervals[1].pix1 - 0.5,
            self.intervals[1].pix2 - 0.5,
            self.intervals[0].pix1 - 0.5,
            self.intervals[0].pix2 - 0.5,
        )



#
# Copyright 2016-2023 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#

import pytest


def _no_bench_func(method, *args, **kwds):
    return method(*args, **kwds)


@pytest.fixture
def benchmark():
    """Mock function used if pytest-benchmark is not installed.

    For actual benchmarking install pytest-benchmark
    """
    return _no_bench_func



#
# Copyright 2015-2023 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#

import sys


def readc(prompt, default=None, valid=None, question_mark=True):
    """Return a single character read from keyboard

    Parameters
    ----------
    prompt : str
        Prompt string.
    default : str
        Default value.
    valid : str
        String providing valid characters. If None, all characters are
        valid (default).
    question_mark : bool
        If True, display question mark after prompt.

    Returns
    -------
    cresult : str
        Read value.

    """

    cresult = None  # Avoid PyCharm warning

    # question mark
    if question_mark:
        cquestion_mark = ' ? '
    else:
        cquestion_mark = ''

    # main loop
    loop = True
    while loop:

        # display prompt
        if default is None:
            print(prompt + cquestion_mark, end='')
            sys.stdout.flush()
        else:
            print(prompt + ' [' + str(default) + ']' + cquestion_mark, end='')
            sys.stdout.flush()

        # read user's input
        cresult = sys.stdin.readline().strip()
        if cresult == '' and default is not None:
            cresult = str(default)

        if len(cresult) == 1:
            # check that all the characters are valid
            loop = False
            if valid is not None:
                for c in cresult:
                    if c not in str(valid):
                        print('*** Error: invalid characters found.')
                        print('*** Valid characters are:', valid)
                        print('*** Try again!')
                        loop = True
        else:
            print('*** Error: invalid string length. Try again!')

    return cresult


def readi(prompt, default=None, minval=None, maxval=None,
          allowed_single_chars=None, question_mark=True):
    """Return integer value read from keyboard

    Parameters
    ----------
    prompt : str
        Prompt string.
    default : integer, str or None
        Default value.
    minval :  integer or None
        Mininum allowed value.
    maxval :  integer or None
        Maximum allowed value.
    allowed_single_chars : str
        String containing allowed valid characters.
    question_mark : bool
        If True, display question mark after prompt.

    Returns
    -------
    result : integer
        Read value.

    """

    return read_value(ftype=int,
                      prompt=prompt,
                      default=default,
                      minval=minval,
                      maxval=maxval,
                      allowed_single_chars=allowed_single_chars,
                      question_mark=question_mark)


def readf(prompt, default=None, minval=None, maxval=None,
          allowed_single_chars=None, question_mark=True):
    """Return integer value read from keyboard

    Parameters
    ----------
    prompt : str
        Prompt string.
    default : float or None
        Default value.
    minval :  float or None
        Mininum allowed value.
    maxval :  float or None
        Maximum allowed value.
    allowed_single_chars : str
        String containing allowed valid characters.
    question_mark : bool
        If True, display question mark after prompt.

    Returns
    -------
    result : float
        Read value.

    """

    return read_value(ftype=float,
                      prompt=prompt,
                      default=default,
                      minval=minval,
                      maxval=maxval,
                      allowed_single_chars=allowed_single_chars,
                      question_mark=question_mark)


def read_value(ftype, prompt, default=None, minval=None, maxval=None,
               allowed_single_chars=None, question_mark=True):
    """Return value read from keyboard

    Parameters
    ----------
    ftype : int() or float()
        Function defining the expected type.
    prompt : str
        Prompt string.
    default : int or None
        Default value.
    minval : int or None
        Mininum allowed value.
    maxval : int or None
        Maximum allowed value.
    allowed_single_chars : str
        String containing allowed valid characters.
    question_mark : bool
        If True, display question mark after prompt.

    Returns
    -------
    result : integer, float or str
        Integer, float of single character.

    """

    # avoid PyCharm warning 'might be referenced before assignment'
    result = None

    # question mark
    if question_mark:
        cquestion_mark = ' ? '
    else:
        cquestion_mark = ''

    # check minimum value
    if minval is not None:
        try:
            iminval = ftype(minval)
        except ValueError:
            raise ValueError("'" + str(minval) + "' cannot " +
                             "be used as an minval in readi()")
    else:
        iminval = None

    # check maximum value
    if maxval is not None:
        try:
            imaxval = ftype(maxval)
        except ValueError:
            raise ValueError("'" + str(maxval) + "' cannot " +
                             "be used as an maxval in readi()")
    else:
        imaxval = None

    # minimum and maximum values
    if minval is None and maxval is None:
        cminmax = ''
    elif minval is None:
        cminmax = ' (number <= ' + str(imaxval) + ')'
    elif maxval is None:
        cminmax = ' (number >= ' + str(iminval) + ')'
    else:
        cminmax = ' (' + str(minval) + ' <= number <= ' + str(maxval) + ')'

    # main loop
    loop = True
    while loop:

        # display prompt
        if default is None:
            print(prompt + cminmax + cquestion_mark, end='')
            sys.stdout.flush()
        else:
            print(prompt + cminmax + ' [' + str(default) + ']' +
                  cquestion_mark, end='')
            sys.stdout.flush()

        # read user's input
        cresult = sys.stdin.readline().strip()
        if cresult == '' and default is not None:
            cresult = str(default)

        # if valid allowed single character, return character
        if len(cresult) == 1:
            if allowed_single_chars is not None:
                if cresult in allowed_single_chars:
                    return cresult

        # convert to ftype value
        try:
            result = ftype(cresult)
        except ValueError:
            print("*** Error: invalid " + str(ftype) + " value. Try again!")
        else:
            # check number is within expected range
            if minval is None and maxval is None:
                loop = False
            elif minval is None:
                if result <= imaxval:
                    loop = False
                else:
                    print("*** Error: number out of range. Try again!")
            elif maxval is None:
                if result >= iminval:
                    loop = False
                else:
                    print("*** Error: number out of range. Try again!")
            else:
                if iminval <= result <= imaxval:
                    loop = False
                else:
                    print("*** Error: number out of range. Try again!")

    return result


def main():

    c = readc("1 Enter character", default='y')
    print('>>>', c)
    c = readc("2 Enter character", default='y', valid='ynx')
    print('>>>', c)
    c = readc("3 Enter character", default='oso')
    print('>>>', c)
    c = readc("4 Enter character", default='oso', valid='yn')
    print('>>>', c)

    i = readi("1 Enter integer", default=8)
    print('>>>', i)
    i = readi("2 Enter integer", minval=0)
    print('>>>', i)
    i = readi("3 Enter integer", default=7, maxval=10)
    print('>>>', i)
    i = readi("4 Enter integer", default=101, minval=5, maxval=10)
    print('>>>', i)

    f = readf("1 Enter float", default=8)
    print('>>>', f)
    f = readf("2 Enter float", minval=0)
    print('>>>', f)
    f = readf("3 Enter float", default=7, maxval=10)
    print('>>>', f)
    f = readf("4 Enter float", default=101, minval=5, maxval=10)
    print('>>>', f)


if __name__ == "__main__":

    main()



#
# Copyright 2015-2023 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#

"""Sequences base class"""


class Sequence(object):
    def __init__(self, instrument, mode):
        self.instrument = instrument
        self.mode = mode

    def setup_instrument(self, instrument):
        pass

    def run(self, **kwds):
        raise NotImplementedError




from ..frame import DataFrameType
from ..dataframe import DataFrame


def test_dataframe_convert_none():

    datatype = DataFrameType()

    assert datatype.convert(None) is None


def test_dataframe_convert_string():

    datatype = DataFrameType()

    obj = 'filename.fits'

    result = datatype.convert(obj)

    assert isinstance(result, DataFrame)
    assert result.filename == obj
    # FIXME: no way of caomparino DataFrame for equality
    # assert result == DataFrame(filename=obj)


def test_dataframe_validate_none():
    datatype = DataFrameType()
    assert datatype.validate(None)



#
# Copyright 2008-2024 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#

"""
Basic tools and classes used to generate recipe modules.

A recipe is a class that complies with the *reduction recipe API*:

 * The class must derive from :class:`numina.core.BaseRecipe`.

"""

import functools
import json
import logging

from astropy.io import fits

from numina.util.jsonencoder import ExtEncoder
from numina.datamodel import DataModel
from .. import __version__
from .recipeinout import RecipeResult as RecipeResultClass
from .recipeinout import RecipeInput as RecipeInputClass
from .metarecipes import RecipeType
from .oresult import ObservationResult, ObservingBlock
from ..exceptions import NoResultFound
from numina.core.taggers import extract_tags_from_obsres, extract_tags_from_img


class BaseRecipe(metaclass=RecipeType):
    """Base class for all instrument recipes

    Parameters
    ----------
    intermediate_results : bool, optional
                           If True, save intermediate results of the Recipe


    Attributes
    ----------

    obresult : ObservationResult, requirement

    logger :
         recipe logger

    """
    RecipeResult = RecipeResultClass
    RecipeInput = RecipeInputClass
    datamodel = DataModel()
    # Recipe own logger
    logger = logging.getLogger('numina.recipes.numina')

    def __new__(cls, *args, **kwargs):
        recipe = super(BaseRecipe, cls).__new__(cls)
        recipe.instrument = kwargs.get('instrument', 'UNKNOWN')
        recipe.mode = kwargs.get('mode', 'UNKNOWN')
        recipe.pipeline = kwargs.get('pipeline', 'default')
        recipe.intermediate_results = kwargs.get('intermediate_results', False)
        recipe.validate_inputs = kwargs.get('validate_inputs', False)
        recipe.validate_results = kwargs.get('validate_results', False)
        recipe.runinfo = cls.create_default_runinfo()
        recipe.runinfo.update(kwargs.get('runinfo', {}))
        recipe.environ = {}
        recipe.__version__ = 1
        recipe.query_options = kwargs.get('query_options', {})
        recipe.configure(**kwargs)
        return recipe

    def __init__(self, *args, **kwargs):
        super(BaseRecipe, self).__init__()
        self.configure(**kwargs)

    def configure(self, **kwds):
        if 'version' in kwds:
            self.__version__ = kwds['version']

        base_kwds = ['instrument', 'mode', 'runinfo', 'intermediate_results',
                     'validate_inputs']
        for kwd in base_kwds:
            if kwd in kwds:
                setattr(self, kwd, kwds[kwd])

    def __setstate__(self, state):
        self.configure(**state)

    @staticmethod
    def create_default_runinfo():
        runinfo = {
            'data_dir': None,
            'results_dir': None,
            'work_dir': None,
            'pipeline': 'default',
            'runner': 'unknown-runner',
            'runner_version': 'unknown-version',
            'taskid': 'unknown-id'
        }
        return runinfo

    @classmethod
    def create_input(cls, *args, **kwds):
        """Pass the result arguments to the RecipeInput constructor"""

        return cls.RecipeInput(*args, **kwds)

    @classmethod
    def create_result(cls, *args, **kwds):
        """Pass the result arguments to the RecipeResult constructor"""

        return cls.RecipeResult(*args, **kwds)

    @classmethod
    def requirements(cls):
        return cls.RecipeInput.stored()

    @classmethod
    def products(cls):
        return cls.RecipeResult.stored()

    @classmethod
    def tag_names(cls):
        return cls.RecipeInput.tag_names()

    def run(self, recipe_input):
        return self.create_result()

    def run_qc(self, recipe_input, recipe_result):
        """Run Quality Control checks."""
        return recipe_result

    def __call__(self, recipe_input):
        """
        Process the result of the observing block with the Recipe.

        Parameters
        ----------
        recipe_input : RecipeInput
                       The input appropriated for the Recipe

        Returns
        -------
        a RecipeResult object or an error
        """

        # Optional input validation
        if self.validate_inputs:
            self.validate_input(recipe_input)

        result = self.run(recipe_input)

        # Update QC in the result
        self.run_qc(recipe_input, result)

        if self.validate_results:
            self.validate_result(result)

        return result

    def validate_input(self, recipe_input):
        """"Validate the input of the recipe"""
        self.logger.debug('start validating input')
        recipe_input.validate()

    def validate_result(self, recipe_result):
        """Validate the result of the recipe"""
        self.logger.debug('start validating result')
        recipe_result.validate()

    def save_intermediate_img(self, img, name):
        """Save intermediate FITS objects."""
        if self.intermediate_results:
            img.writeto(name, overwrite=True)

    def save_intermediate_array(self, array, name):
        """Save intermediate array object as FITS."""
        if self.intermediate_results:
            fits.writeto(name, array, overwrite=True)

    def save_structured_as_json(self, structured, name):
        if self.intermediate_results:
            if hasattr(structured, '__getstate__'):
                state = structured.__getstate__()
            elif isinstance(structured, dict):
                state = structured
            else:
                state = structured.__dict__

            with open(name, 'w') as fd:
                json.dump(state, fd, indent=2, cls=ExtEncoder)

    def set_base_headers(self, hdr):
        """Set metadata in FITS headers."""
        hdr['NUMXVER'] = (__version__, 'Numina package version')
        hdr['NUMRNAM'] = (self.__class__.__name__, 'Numina recipe name')
        hdr['NUMRVER'] = (self.__version__, 'Numina recipe version')
        return hdr

    def build_recipe_input(self, ob, dal):
        """Build a RecipeInput object."""
        result = {}
        # We have to decide if the ob input
        # is a plain description (ObservingBlock)
        # or if it contains the nested results (Obsres)
        #
        # it has to contain the tags corresponding to the observing modes...
        ob_query_skip = False
        ob_query_field = 'obresult'

        if isinstance(ob, ObservingBlock):
            import numina.types.obsresult as obtype
            # We have to build an Obsres
            for key, req in self.requirements().items():
                if isinstance(req.type, obtype.ObservationResultType):
                    ob_query_field = key
                    ob_query_skip = True
                    query_option = self.query_options.get(key)

                    # print('req for ob is named', key, query_option)
                    new_or = ObservationResult()
                    new_or.__dict__ = ob.__dict__
                    obsres = req.query(dal, new_or, options=query_option)
                    break
            else:
                # nothing to do
                obsres = ob
        else:
            obsres = ob

        qfields = self.tag_names()
        self.logger.debug(
            'running recipe tagger with query fields: %s', qfields)
        if qfields:
            obsres.tags = extract_tags_from_obsres(obsres, qfields, self.datamodel, strict=True)
        else:
            obsres.tags = {}
        self.logger.debug('obsres tags are: %s', obsres.tags)

        for key, req in self.requirements().items():

            try:
                query_option = self.query_options.get(key)
                if key == ob_query_field and ob_query_skip:
                    result[key] = obsres
                else:
                    result[key] = req.query(dal, obsres, options=query_option)
            except NoResultFound as notfound:
                req.on_query_not_found(notfound)

        return self.create_input(**result)

    def extract_tags_from_ref(self, ref, tag_keys, base=None):
        return extract_tags_from_img(ref, tag_keys, self.datamodel, base=base)

    def get_filters(self, imgtypes, getters):
        import collections.abc

        used_getters = []
        for rtype, getter in zip(imgtypes, getters):
            self.logger.debug('get_filters, %s  %s', rtype, getter)
            if rtype is None:
                # Unconditional
                if isinstance(getter, collections.abc.Iterable):
                    used_getters.extend(getter)
                else:
                    used_getters.append(getter)
            else:
                # Search
                for key, val in self.RecipeInput.stored().items():
                    if isinstance(val.type, rtype):
                        if isinstance(getter, collections.abc.Iterable):
                            used_getters.extend(getter)
                        else:
                            used_getters.append(getter)
                        break
                else:
                    pass
        return used_getters

    def init_filters_generic(self, rinput, getters_seq, ins):
        import numina.ext.gtc
        import numina.util.flow as flowmod

        # with BPM, bias, dark, flat and sky
        if numina.ext.gtc.check_gtc():
            self.logger.debug('running in GTC environment')
        else:
            self.logger.debug('running outside of GTC environment')

        meta = self.gather_info(rinput)
        self.logger.debug('obresult info')
        for entry in meta['obresult']:
            self.logger.debug('frame info is %s', entry)

        reduction_flows = []
        for getters in getters_seq:
            correctors = [getter(rinput, meta, ins, self.datamodel)
                          for getter in getters]
            reduction_flow = flowmod.SerialFlow(correctors)
            reduction_flows.append(reduction_flow)

        return reduction_flows

    def init_filters(self, rinput, ins=None):
        # Defined to be overloaded
        if ins is None:
            ins = rinput.obresult.configuration

        getters_seq = self.create_getters()
        return self.init_filters_generic(rinput, getters_seq, ins)

    def create_getters(self):
        return []

    def gather_info(self, recipeinput):
        # Requires collaboration from datamodel
        from numina.core import DataFrame

        klass = recipeinput.__class__
        metadata = {}
        for key in klass.stored():
            val = getattr(recipeinput, key)
            if isinstance(val, DataFrame):
                metadata[key] = self.datamodel.gather_info_dframe(val)
            # elif isinstance(val, ObservationResult):
            elif hasattr(val, 'frames'):
                metadata[key] = self.datamodel.gather_info_oresult(val)
            else:
                pass
        return metadata


def timeit(method):
    """Decorator to measure the time used by the recipe"""

    import datetime

    @functools.wraps(method)
    def timed_method(self, rinput):

        time_start = datetime.datetime.now(datetime.timezone.utc)
        result = method(self, rinput)
        time_end = datetime.datetime.now(datetime.timezone.utc)
        result.time_it(time_start, time_end)
        self.logger.info('total time measured')
        return result

    return timed_method



import pytest


def test_mod_proxy1(pattri_ins, pattri_header):
    pattri_ins.configure_with_header(pattri_header)
    assert pattri_ins.get_property("key2") == "A"


def test_mod_proxy2(pattri_ins, pattri_header):
    pattri_ins.configure_with_header(pattri_header)
    assert pattri_ins.get_property("key2") == "A"
    pattri_ins.key2 = "B"
    assert pattri_ins.get_property("key2") == "B"


def test_mod_proxy3(pattri_ins, pattri_header):
    pattri_ins.configure_with_header(pattri_header)
    assert pattri_ins.get_property("key3") == "B"


def test_mod_limit1(pattri_ins):
    assert pattri_ins.get_property("key1") == 0
    assert pattri_ins.get_property("key11") == 23
    assert pattri_ins.get_property("key12") == 12
    assert pattri_ins.get_property("key13") == 8
    assert pattri_ins.get_property("key14") == 0


def test_mod_limit2(pattri_ins):
    pattri_ins.key1 = 17
    assert pattri_ins.get_property("key1") == 17
    pattri_ins.key11 = 22
    assert pattri_ins.get_property("key11") == 22


def test_mod_limit3(pattri_ins):
    with pytest.raises(ValueError):
        pattri_ins.key11 = 50



#
# Copyright 2015-2023 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#

"""Simple monocromatic simulation"""

import json


class RunCounter(object):
    """Run number counter"""

    def __init__(self, template, last=1):
        self.template = template
        self.last = last

    def runstring(self):
        """Return the run number and the file name."""
        cfile = self.template % self.last
        self.last += 1
        return cfile


class PersistentRunCounter(RunCounter):
    """Persistent run number counter"""

    def __init__(self, template, last=1, pstore='index.json',):

        last = self.load(pstore, last)

        super(PersistentRunCounter, self).__init__(template, last)

        self.pstore = pstore

    def store(self):
        with open(self.pstore, 'w') as pkl_file:
            json.dump(self.last, pkl_file)

    @staticmethod
    def load(pstore, last):
        file_exists = True

        try:
            with open(pstore, 'rb') as pkl_file:
                last = json.load(pkl_file)
        except IOError:
            file_exists = False

        if not file_exists:
            with open(pstore, 'wb') as pkl_file:
                json.dump(last, pkl_file)

        return last

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.store()


def extract(header, meta, path, key, selector=None, default=None):
    m = meta
    if selector is None:
        def selector(x): return x  # noqa: E731
    try:
        for part in path:
            m = m[part]
        header[key] = selector(m)
    except KeyError:
        # Keyword missing
        if default is not None:
            header[key] = default


def extractm(meta, path, selector=None):
    m = meta
    if selector is None:
        def selector(x): return x  # noqa: E731
    for part in path:
        m = m[part]
    return selector(m)


if __name__ == '__main__':

    with PersistentRunCounter('r00%04d') as p:
        for i in range(10):
            print(p.runstring())

    with PersistentRunCounter('r00%04d') as p:
        for i in range(10):
            print(p.runstring())



#
# Copyright 2008-2022 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#

"""Background estimation

    Background estimation following Costa 1992, Bertin & Arnouts 1996
"""


import numpy
import scipy.ndimage as ndimage


def _interpolation(z, sx, sy, mx, my):
    # Spline to original size
    x, y = numpy.ogrid[-1:1:complex(0, mx), -1:1:complex(0, my)]
    newx, newy = numpy.mgrid[-1:1:complex(0, sx), -1:1:complex(0, sy)]

    x0 = x[0, 0]
    y0 = y[0, 0]
    dx = x[1, 0] - x0
    dy = y[0, 1] - y0
    ivals = (newx - x0)/dx
    jvals = (newy - y0)/dy
    coords = numpy.array([ivals, jvals])
    newf = ndimage.map_coordinates(z, coords)
    return newf


def background_estimator(bdata):
    """Estimate the background in a 2D array"""

    crowded = False

    std = numpy.std(bdata)
    std0 = std
    mean = bdata.mean()
    while True:
        prep = len(bdata)
        numpy.clip(bdata, mean - 3 * std, mean + 3 * std, out=bdata)
        if prep == len(bdata):
            if std < 0.8 * std0:
                crowded = True
            break
        std = numpy.std(bdata)
        mean = bdata.mean()

    if crowded:
        median = numpy.median(bdata)
        mean = bdata.mean()
        std = bdata.std()
        return 2.5 * median - 1.5 * mean, std

    return bdata.mean(), bdata.std()


def create_background_map(data, bsx, bsy):
    """Create a background map with a given mesh size"""
    sx, sy = data.shape
    mx = sx // bsx
    my = sy // bsy
    comp = []
    rms = []
    # Rows
    sp = numpy.split(data, numpy.arange(bsx, sx, bsx), axis=0)
    for s in sp:
        # Columns
        rp = numpy.split(s, numpy.arange(bsy, sy, bsy), axis=1)
        for r in rp:
            b, r = background_estimator(r)
            comp.append(b)
            rms.append(r)

    # Reconstructed image
    z = numpy.array(comp)
    z.shape = (mx, my)
    # median filter
    ndimage.median_filter(z, size=(3, 3), output=z)

    # Interpolate to the original size
    new = _interpolation(z, sx, sy, mx, my)

    # Interpolate the rms
    z = numpy.array(rms)
    z.shape = (mx, my)
    nrms = _interpolation(z, sx, sy, mx, my)

    return new, nrms




import numina.core
import numina.core.query
import numina.dal.stored
import numina.exceptions
import numina.types.datatype as dt
import numina.tests.simpleobj
from numina.types.structured import BaseStructuredCalibration
from numina.types.multitype import MultiType
from ..dataholders import Requirement


class One(BaseStructuredCalibration):
    pass


class Other(BaseStructuredCalibration):
    pass


class Dal(object):
    def search_parameter(self, name, stype, obsres, options=None):
        if name == 'req5':
            return numina.dal.stored.StoredParameter(content=[1, 300, 12])
        else:
            return numina.dal.stored.StoredParameter(content=1)

    def search_product(self, name, stype, obsres, options=None):
        if name == 'req4':
            if stype.name() == 'One' and obsres.tags == {'a': 1, 'b': 2}:
                return numina.dal.stored.StoredProduct(id=1, content=11, tags={})
            elif stype.name() == 'One' and obsres.tags == {'a': 2, 'b': 1}:
                return numina.dal.stored.StoredProduct(id=10, content=12, tags={})
            elif stype.name() == 'Other':
                return numina.dal.stored.StoredProduct(id=100, content=24, tags={})
            else:
                raise numina.exceptions.NoResultFound(
                    f'{stype.name()} not found')
        else:
            return numina.dal.stored.StoredProduct(id=1, content=2, tags={})


def test_list_of():

    req1 = Requirement(dt.ListOfType(
        One), description="ListOf", destination="req1")
    req2 = Requirement(dt.PlainPythonType(ref=100),
                       description="simple", destination="req2")
    req3 = Requirement(MultiType(One, Other),
                       description="simple", destination="req3")
    req4 = Requirement(dt.ListOfType(MultiType(One, Other)),
                       description="simple", destination="req4")
    req5 = Requirement(dt.ListOfType(dt.PlainPythonType(
        ref=100)), description="simple", destination="req5")

    dal = Dal()

    obsres = numina.core.ObservationResult()
    obsres.tags = [{'a': 1, 'b': 2}, {'a': 2, 'b': 1}, {'a': 2, 'b': 2}]

    assert obsres.tags == [{'a': 1, 'b': 2},
                           {'a': 2, 'b': 1}, {'a': 2, 'b': 2}]
    res1 = req1.query(dal, obsres)
    assert obsres.tags == [{'a': 1, 'b': 2},
                           {'a': 2, 'b': 1}, {'a': 2, 'b': 2}]
    res2 = req2.query(dal, obsres)
    assert obsres.tags == [{'a': 1, 'b': 2},
                           {'a': 2, 'b': 1}, {'a': 2, 'b': 2}]
    res3 = req3.query(dal, obsres)
    assert obsres.tags == [{'a': 1, 'b': 2},
                           {'a': 2, 'b': 1}, {'a': 2, 'b': 2}]
    res4 = req4.query(dal, obsres)
    assert obsres.tags == [{'a': 1, 'b': 2},
                           {'a': 2, 'b': 1}, {'a': 2, 'b': 2}]
    res5 = req5.query(dal, obsres)
    assert obsres.tags == [{'a': 1, 'b': 2},
                           {'a': 2, 'b': 1}, {'a': 2, 'b': 2}]

    assert res1 == [2, 2, 2]
    assert res2 == 1
    assert res3 == 2
    assert res4 == [11, 12, 24]
    assert res5 == [1, 300, 12]



#
# Copyright 2014-2015 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#

"""DAL Mock class"""


from numina.core.utils import AlwaysSuccessRecipe
from .stored import StoredProduct, StoredParameter
from .stored import ObservingBlock
from .absdal import AbsDAL


class MockDAL(AbsDAL):

    def search_oblock_from_id(self, objid):
        ob = ObservingBlock(1, 'TEST_INSTRUMENT', 'TEST_MODE', [], [], None)
        return ob

    # Implemented in base class
    # def obsres_from_oblock_id(self, obsid):

    def search_recipe_from_ob(self, ob, pipeline='default'):

        return AlwaysSuccessRecipe

    def search_prod_type_tags(self, ins, type, tags, pipeline):
        '''Returns the first coincidence...'''
        return StoredProduct(id=100, content='null.fits', tags={})

    def search_prod_req_tags(self, req, ins, tags, pipeline):
        return self.search_prod_type_tags(req.type, ins, tags, pipeline)

    # Implemented in base class
    # def obsres_from_proc_oblock_id(self, instrument, child_id)

    def search_prod_obsid(self, ins, obsid, pipeline):
        return StoredProduct(id=100, content='null.fits',
                             tags={'readmode': 'a'})

    def search_param_req_tags(self, req, instrument, mode, tags, pipeline):
        return StoredParameter(content='parameter')



#
# Copyright 2024-2025 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#

from astropy.io import fits
from astropy.units import Quantity
import astropy.units as u
from joblib import Parallel, delayed
import json
import numpy as np
import os
import pprint
import time
import yaml

from numina.tools.ctext import ctext

from .define_3d_wcs import get_wvparam_from_wcs3d
from .define_3d_wcs import wcs_to_header_using_cd_keywords
from .display_skycalc import display_skycalc
from .compute_image2d_rss_from_detector_method1 import compute_image2d_rss_from_detector_method1
from .compute_image3d_ifu_from_rss_method1 import compute_image3d_ifu_from_rss_method1
from .generate_image2d_method0_ifu import generate_image2d_method0_ifu
from .generate_geometry_for_scene_block import generate_geometry_for_scene_block
from .generate_image3d_method0_ifu import generate_image3d_method0_ifu
from .generate_spectrum_for_scene_block import generate_spectrum_for_scene_block
from .load_atmosphere_transmission_curve import load_atmosphere_transmission_curve
from .raise_valueerror import raise_ValueError
from .save_image2d_detector_method0 import save_image2d_detector_method0
from .save_image2d_rss import save_image2d_rss
from .set_wavelength_unit_and_range import set_wavelength_unit_and_range
from .update_image2d_rss_detector_method0 import update_image2d_rss_detector_method0


pp = pprint.PrettyPrinter(indent=1, sort_dicts=False)


def ifu_simulator(wcs3d, header_keys,
                  naxis1_detector, naxis2_detector, nslices,
                  noversampling_whitelight,
                  scene_fname,
                  flux_factor,
                  seeing_fwhm_arcsec, seeing_psf,
                  instrument_pa,
                  airmass,
                  parallactic_angle,
                  flatpix2pix,
                  atmosphere_transmission,
                  bias,
                  rnoise,
                  spectral_blurring_pixel,
                  faux_dict,
                  rng,
                  noparallel_computation,
                  prefix_intermediate_fits,
                  stop_after_ifu_3D_method0=False,
                  verbose=False, instname=None, subtitle=None, plots=False):
    """IFU simulator.

    Parameters
    ----------
    wcs3d : `~astropy.wcs.wcs.WCS`
        WCS of the data cube.
    header_keys : `~astropy.io.fits.header.Header`
        FITS header with additional keywords to be merged together with
        the WCS information.
    naxis1_detector : `~astropy.units.Quantity`
        Detector NAXIS1, dispersion direction.
    naxis2_detector : `~astropy.units.Quantity`
        Detector NAXIS2, spatial direction (slices).
    nslices : int
        Number of IFU slices.
    noversampling_whitelight : int
        Oversampling factor (integer number) to generate the white
        light image.
    scene_fname : str
        YAML scene file name.
    flux_factor : float
        Flux factor to be applied to the number of photons defined
        in the scene file.
    seeing_fwhm_arcsec : `~astropy.units.Quantity`
        Seeing FWHM.
    seeing_psf : str
        Seeing PSF.
    instrument_pa : `~astropy.units.Quantity`
        Instrument position angle.
    airmass : float
        Airmass.
    parallactic_angle : `~astropy.units.Quantity`
        Parallactic angle. This number must be within the range
        [-90,+90] deg.
    flatpix2pix : str
        String indicating whether a pixel-to-pixel flatfield is
        applied or not. Two possible values:
        - 'default': use default flatfield defined in 'faux_dict'
        - 'none': do not apply flatfield
    atmosphere_transmission : str
        String indicating whether the atmosphere transmission of
        the atmosphere is applied or not. Two possible values are:
        - 'default': use default curve defined in 'faux_dict'
        - 'none': do not apply atmosphere transmission
    bias : `~astropy.units.Quantity`
        Bias level (in ADU).
    rnoise : `~astropy.units.Quantity`
        Readout noise standard deviation (in ADU). Assumed to be
        Gaussian.
    spectral_blurring_pixel : `~astropy.units.Quantity`
        Spectral blurring (in pixels) to be introduced when generating
        the initial RSS image from the original 3D data cube.
    faux_dict : Python dictionary
        File names of auxiliary files:
        - skycalc: table with SKYCALC Sky Model Calculator predictions
        - flatpix2pix: pixel-to-pixel flat field
        - model_ifu2detector: 2D polynomial transformation
          x_ifu, y_ify, wavelength -> x_detector, y_detector
    rng : `~numpy.random._generator.Generator`
        Random number generator.
    noparallel_computation : bool
        It True, skip use of parallel processing.
    prefix_intermediate_fits : str
        Prefix for output intermediate FITS files. If the length of
        this string is 0, no output is generated.
    stop_after_ifu_3D_method0 : bool
        If True, stop execution after generating ifu_3D_method0 image.
    verbose : bool
        If True, display additional information.
    instname : str or None
        Instrument name.
    subtitle : str or None
        Plot subtitle.
    plots : bool
        If True, plot intermediate results.

    Returns
    -------
    """

    if verbose:
        print(' ')
        for item in faux_dict:
            print(ctext(f'- Required file for item {item}:\n  {faux_dict[item]}', faint=True))

    if plots:
        # display SKYCALC predictions for sky radiance and transmission
        display_skycalc(faux_skycalc=faux_dict['skycalc'])

    # spatial IFU limits
    naxis1_ifu = Quantity(value=wcs3d.array_shape[2], unit=u.pix, dtype=int)
    naxis2_ifu = Quantity(value=wcs3d.array_shape[1], unit=u.pix, dtype=int)
    min_x_ifu = 0.5 * u.pix
    max_x_ifu = naxis1_ifu + 0.5 * u.pix
    min_y_ifu = 0.5 * u.pix
    max_y_ifu = naxis2_ifu + 0.5 * u.pix

    # wavelength limits
    wv_cunit1, wv_crpix1, wv_crval1, wv_cdelt1 = get_wvparam_from_wcs3d(wcs3d)
    wmin = wv_crval1 + (0.5 * u.pix - wv_crpix1) * wv_cdelt1
    wmax = wmin + naxis1_detector * wv_cdelt1
    reference_wave_vacuum_differential_refraction = (wmin + wmax) / 2

    # load atmosphere transmission curve
    wave_transmission, curve_transmission = load_atmosphere_transmission_curve(
        atmosphere_transmission=atmosphere_transmission,
        wmin=wmin,
        wmax=wmax,
        wv_cunit1=wv_cunit1,
        faux_dict=faux_dict,
        verbose=verbose
    )

    required_keys_in_scene_block = {
        'scene_block_name',
        'spectrum',
        'geometry',
        'nphotons',
        'wavelength_sampling',             # default: random
        'apply_atmosphere_transmission',   # default: True
        'apply_seeing',                    # default: True
        'render'                           # default: True
    }

    nphotons_all = 0
    simulated_wave_all = None
    simulated_x_ifu_all = None
    simulated_y_ifu_all = None

    # main loop (rendering of scene blocks)
    with open(scene_fname, 'rt') as fstream:
        scene_dict = yaml.safe_load_all(fstream)
        for scene_block in scene_dict:
            if 'scene_block_name' not in scene_block.keys():
                if verbose:
                    print(' ')
                print(ctext(f'ERROR while processing {scene_fname}', fg='red'))
                raise_ValueError('key scene_block_name not found!')
            scene_block_name = scene_block['scene_block_name']
            print(ctext(f'\n* Processing: {scene_block_name}', fg='green'))
            # insert default values for keys not provided
            if 'wavelength_sampling' not in scene_block.keys():
                scene_block['wavelength_sampling'] = 'random'
                print(ctext(f"WARNING: asumming {scene_block['wavelength_sampling']=}", fg='cyan'))
            if 'render' not in scene_block.keys():
                scene_block['render'] = True
                print(ctext(f"WARNING: asumming {scene_block['render']=}", fg='cyan'))
            if 'apply_atmosphere_transmission' not in scene_block.keys():
                scene_block['apply_atmosphere_transmission'] = True
                print(ctext(f"WARNING: asumming {scene_block['apply_atmosphere_transmission']=}", fg='cyan'))
            if 'apply_seeing' not in scene_block.keys():
                scene_block['apply_seeing'] = True
                print(ctext(f"WARNING: asumming {scene_block['apply_seeing']=}", fg='cyan'))
            scene_block_keys = set(scene_block.keys())
            # insert default values
            if scene_block_keys != required_keys_in_scene_block:
                print(ctext(f'ERROR while processing: {scene_fname}', fg='red'))
                print(ctext('expected keys..: ', fg='blue') + f'{required_keys_in_scene_block}')
                print(ctext('keys found.....: ', fg='blue') + f'{scene_block_keys}')
                list_unexpected_keys = list(scene_block_keys.difference(required_keys_in_scene_block))
                if len(list_unexpected_keys) > 0:
                    print(ctext('unexpected keys: ', fg='red') + f'{list_unexpected_keys}')
                list_missing_keys = list(required_keys_in_scene_block.difference(scene_block_keys))
                if len(list_missing_keys) > 0:
                    print(ctext('missing keys...: ', fg='red') + f'{list_missing_keys}')
                pp.pprint(scene_block)
                raise_ValueError(f'Invalid format in file: {scene_fname}')
            if verbose:
                pp.pprint(scene_block)

            nphotons = int(float(scene_block['nphotons']) * flux_factor)
            wavelength_sampling = scene_block['wavelength_sampling']
            if wavelength_sampling not in ['random', 'fixed']:
                raise_ValueError(f'Unexpected {wavelength_sampling=}')
            apply_atmosphere_transmission = scene_block['apply_atmosphere_transmission']
            if atmosphere_transmission == "none" and apply_atmosphere_transmission:
                print(ctext(f'WARNING: {apply_atmosphere_transmission=} when {atmosphere_transmission=}', fg='cyan'))
                print(f'{atmosphere_transmission=} overrides {apply_atmosphere_transmission=}')
                print('The atmosphere transmission will not be applied!')
                apply_atmosphere_transmission = False
            apply_seeing = scene_block['apply_seeing']
            if apply_seeing:
                if seeing_fwhm_arcsec.value < 0:
                    raise_ValueError(f'Unexpected {seeing_fwhm_arcsec=}')
                elif seeing_fwhm_arcsec == 0:
                    print(ctext(f'WARNING: {apply_seeing=} when {seeing_fwhm_arcsec=}', fg='cyan'))
                    print('Seeing effect will not be applied!')
                    apply_seeing = False
            render = scene_block['render']
            if nphotons > 0 and render:
                # set wavelength unit and range
                wave_unit, wave_min, wave_max = set_wavelength_unit_and_range(
                    scene_fname=scene_fname,
                    scene_block=scene_block,
                    wmin=wmin,
                    wmax=wmax,
                    verbose=verbose
                )
                # generate spectrum
                simulated_wave = generate_spectrum_for_scene_block(
                    scene_fname=scene_fname,
                    scene_block=scene_block,
                    faux_dict=faux_dict,
                    wave_unit=wave_unit,
                    wave_min=wave_min,
                    wave_max=wave_max,
                    nphotons=nphotons,
                    wavelength_sampling=wavelength_sampling,
                    apply_atmosphere_transmission=apply_atmosphere_transmission,
                    wave_transmission=wave_transmission,
                    curve_transmission=curve_transmission,
                    rng=rng,
                    naxis1_detector=naxis1_detector,
                    verbose=verbose,
                    plots=plots
                )
                # convert to default wavelength_unit
                simulated_wave = simulated_wave.to(wv_cunit1)
                # distribute photons in the IFU focal plane
                simulated_x_ifu, simulated_y_ifu = generate_geometry_for_scene_block(
                    scene_fname=scene_fname,
                    scene_block=scene_block,
                    nphotons=nphotons,
                    apply_seeing=apply_seeing,
                    seeing_fwhm_arcsec=seeing_fwhm_arcsec,
                    seeing_psf=seeing_psf,
                    airmass=airmass,
                    parallactic_angle=parallactic_angle,
                    reference_wave_vacuum_differential_refraction=reference_wave_vacuum_differential_refraction,
                    simulated_wave=simulated_wave,
                    instrument_pa=instrument_pa,
                    wcs3d=wcs3d,
                    min_x_ifu=min_x_ifu,
                    max_x_ifu=max_x_ifu,
                    min_y_ifu=min_y_ifu,
                    max_y_ifu=max_y_ifu,
                    rng=rng,
                    verbose=verbose,
                    plots=plots
                )
                # store all simulated photons
                if nphotons_all == 0:
                    simulated_wave_all = simulated_wave
                    simulated_x_ifu_all = simulated_x_ifu
                    simulated_y_ifu_all = simulated_y_ifu
                else:
                    simulated_wave_all = np.concatenate((simulated_wave_all, simulated_wave))
                    simulated_x_ifu_all = np.concatenate((simulated_x_ifu_all, simulated_x_ifu))
                    simulated_y_ifu_all = np.concatenate((simulated_y_ifu_all, simulated_y_ifu))
                # ---
                # update nphotons
                if verbose:
                    print(ctext(f'--> {nphotons} photons simulated', fg='blue'))
                if nphotons_all == 0:
                    nphotons_all = nphotons
                else:
                    nphotons_all += nphotons
                if len({nphotons_all,
                        len(simulated_wave_all),
                        len(simulated_x_ifu_all),
                        len(simulated_y_ifu_all)
                        }) != 1:
                    print(ctext('ERROR: check the following numbers:', fg='red'))
                    print(f'{nphotons_all=}')
                    print(f'{len(simulated_wave_all)=}')
                    print(f'{len(simulated_x_ifu_all)=}')
                    print(f'{len(simulated_y_ifu_all)=}')
                    raise_ValueError('Unexpected differences found in the previous numbers')
            else:
                if verbose:
                    if nphotons == 0:
                        print(ctext('WARNING -> nphotons: 0', fg='cyan'))
                    else:
                        print(ctext('WARNING -> render: False', fg='cyan'))

    # filter simulated photons to keep only those that fall within
    # the IFU field of view and within the expected spectral range
    # (note that this step also removes simulated photons with
    # negative wavelength value corresponding to those absorbed by
    # the atmosphere when applying the transmission curve)
    textwidth_nphotons_number = len(str(nphotons_all))
    if verbose:
        print('\nFiltering photons within IFU field of view and spectral range...')
        print(f'Initial number of simulated photons: {nphotons_all:>{textwidth_nphotons_number}}')
    cond1 = simulated_x_ifu_all >= min_x_ifu
    cond2 = simulated_x_ifu_all <= max_x_ifu
    cond3 = simulated_y_ifu_all >= min_y_ifu
    cond4 = simulated_y_ifu_all <= max_y_ifu
    cond5 = simulated_wave_all >= wmin
    cond6 = simulated_wave_all <= wmax
    iok = np.where(cond1 & cond2 & cond3 & cond4 & cond5 & cond6)[0]

    if len(iok) == 0:
        print(ctext(f'Final number of simulated photons..: {len(iok):>{textwidth_nphotons_number}}', fg='red'))
        raise SystemExit

    if len(iok) < nphotons_all:
        simulated_x_ifu_all = simulated_x_ifu_all[iok]
        simulated_y_ifu_all = simulated_y_ifu_all[iok]
        simulated_wave_all = simulated_wave_all[iok]
        nphotons_all = len(iok)
    if verbose:
        print(ctext(f'Final number of simulated photons..: {nphotons_all:>{textwidth_nphotons_number}}', fg='blue'))

    # ---------------------------------------------------------------
    # compute image2d IFU, white image, with and without oversampling
    # ---------------------------------------------------------------
    if verbose:
        print(ctext('\n* Computing image2d IFU (method 0) with and without oversampling', fg='green'))
    for noversampling in {noversampling_whitelight, 1}:
        generate_image2d_method0_ifu(
            wcs3d=wcs3d,
            header_keys=header_keys,
            noversampling_whitelight=noversampling,
            simulated_x_ifu_all=simulated_x_ifu_all,
            simulated_y_ifu_all=simulated_y_ifu_all,
            prefix_intermediate_fits=prefix_intermediate_fits,
            instname=instname,
            subtitle=subtitle,
            scene=scene_fname,
            plots=plots
        )

    # ----------------------------
    # compute image3d IFU, method0
    # ----------------------------
    if verbose:
        print(ctext('\n* Computing image3d IFU (method 0)', fg='green'))
    bins_x_ifu = (0.5 + np.arange(naxis1_ifu.value + 1)) * u.pix
    bins_y_ifu = (0.5 + np.arange(naxis2_ifu.value + 1)) * u.pix
    bins_wave = wv_crval1 + \
        ((np.arange(naxis2_detector.value + 1) + 1) * u.pix - wv_crpix1) * wv_cdelt1 \
        - 0.5 * u.pix * wv_cdelt1
    generate_image3d_method0_ifu(
        wcs3d=wcs3d,
        header_keys=header_keys,
        simulated_x_ifu_all=simulated_x_ifu_all,
        simulated_y_ifu_all=simulated_y_ifu_all,
        simulated_wave_all=simulated_wave_all,
        bins_x_ifu=bins_x_ifu,
        bins_y_ifu=bins_y_ifu,
        bins_wave=bins_wave,
        prefix_intermediate_fits=prefix_intermediate_fits
    )

    if stop_after_ifu_3D_method0:
        raise SystemExit(f'Program stopped because {stop_after_ifu_3D_method0=}')

    # --------------------------------------------
    # compute image2d RSS and in detector, method0
    # --------------------------------------------
    if verbose:
        print(ctext('\n* Computing image2d RSS and detector (method 0)', fg='green'))
    bins_x_detector = np.linspace(start=0.5, stop=naxis1_detector.value + 0.5, num=naxis1_detector.value + 1)
    bins_y_detector = np.linspace(start=0.5, stop=naxis2_detector.value + 0.5, num=naxis2_detector.value + 1)

    # read ifu2detector transformations
    dict_ifu2detector = json.loads(open(faux_dict['model_ifu2detector'], mode='rt').read())

    # additional degradation in the spectral direction
    # (in units of detector pixels)
    if spectral_blurring_pixel.unit != u.pix:
        raise_ValueError(f'Unexpected unit for {spectral_blurring_pixel.unit=}')
    extra_degradation_spectral_direction = rng.normal(
        loc=0.0,
        scale=spectral_blurring_pixel.value,
        size=nphotons_all
    ) * u.pix

    # initialize images
    image2d_rss_method0 = np.zeros((naxis1_ifu.value * nslices, naxis1_detector.value))
    image2d_detector_method0 = np.zeros((naxis2_detector.value, naxis1_detector.value))

    # update images
    # (accelerate computation using joblib.Parallel)
    t0 = time.time()
    if noparallel_computation:
        # explicit loop in slices
        for islice in range(nslices):
            if verbose:
                print(f'{islice=}')
            update_image2d_rss_detector_method0(
                islice=islice,
                simulated_x_ifu_all=simulated_x_ifu_all,
                simulated_y_ifu_all=simulated_y_ifu_all,
                simulated_wave_all=simulated_wave_all,
                naxis1_ifu=naxis1_ifu,
                bins_x_ifu=bins_x_ifu,
                bins_wave=bins_wave,
                bins_x_detector=bins_x_detector,
                bins_y_detector=bins_y_detector,
                wv_cdelt1=wv_cdelt1,
                extra_degradation_spectral_direction=extra_degradation_spectral_direction,
                dict_ifu2detector=dict_ifu2detector,
                image2d_rss_method0=image2d_rss_method0,
                image2d_detector_method0=image2d_detector_method0
            )
    else:
        Parallel(n_jobs=-1, prefer="threads")(
            delayed(update_image2d_rss_detector_method0)(
                islice=islice,
                simulated_x_ifu_all=simulated_x_ifu_all,
                simulated_y_ifu_all=simulated_y_ifu_all,
                simulated_wave_all=simulated_wave_all,
                naxis1_ifu=naxis1_ifu,
                bins_x_ifu=bins_x_ifu,
                bins_wave=bins_wave,
                bins_x_detector=bins_x_detector,
                bins_y_detector=bins_y_detector,
                wv_cdelt1=wv_cdelt1,
                extra_degradation_spectral_direction=extra_degradation_spectral_direction,
                dict_ifu2detector=dict_ifu2detector,
                image2d_rss_method0=image2d_rss_method0,
                image2d_detector_method0=image2d_detector_method0
            ) for islice in range(nslices))
    t1 = time.time()
    if verbose:
        print(f'Delta time: {t1 - t0}')

    # save RSS image (note that the flatfield effect is not included!)
    save_image2d_rss(
        wcs3d=wcs3d,
        header_keys=header_keys,
        image2d_rss=image2d_rss_method0,
        method=0,
        prefix_intermediate_fits=prefix_intermediate_fits,
        bitpix=16
    )

    # apply flatpix2pix to detector image
    if flatpix2pix not in ['default', 'none']:
        raise_ValueError(f'Invalid {flatpix2pix=}')
    if flatpix2pix == 'default':
        infile = faux_dict['flatpix2pix']
        with fits.open(infile) as hdul:
            image2d_flatpix2pix = hdul[0].data
        if np.min(image2d_flatpix2pix) <= 0:
            print(f'- minimum flatpix2pix value: {np.min(image2d_flatpix2pix)}')
            raise_ValueError('Unexpected signal in flatpix2pix <= 0')
        naxis2_flatpix2pix, naxis1_flatpix2pix = image2d_flatpix2pix.shape
        naxis1_flatpix2pix *= u.pix
        naxis2_flatpix2pix *= u.pix
        if (naxis1_flatpix2pix != naxis1_detector) or (naxis2_flatpix2pix != naxis2_detector):
            raise_ValueError(f'Unexpected flatpix2pix shape: naxis1={naxis1_flatpix2pix}, naxis2={naxis2_flatpix2pix}')
        if verbose:
            print(f'Applying flatpix2pix: {os.path.basename(infile)} to detector image')
            print(f'- minimum flatpix2pix value: {np.min(image2d_flatpix2pix):.6f}')
            print(f'- maximum flatpix2pix value: {np.max(image2d_flatpix2pix):.6f}')
        image2d_detector_method0 /= image2d_flatpix2pix
    else:
        if verbose:
            print('Skipping applying flatpix2pix')

    # apply bias to detector image
    if bias.value > 0:
        image2d_detector_method0 += bias.value
        if verbose:
            print(f'Applying {bias=} to detector image')
    else:
        if verbose:
            print('Skipping adding bias')

    # apply Gaussian readout noise to detector image
    if rnoise.value > 0:
        if verbose:
            print(f'Applying Gaussian {rnoise=} to detector image')
        ntot_pixels = naxis1_detector.value * naxis2_detector.value
        image2d_rnoise_flatten = rng.normal(loc=0.0, scale=rnoise.value, size=ntot_pixels)
        image2d_detector_method0 += image2d_rnoise_flatten.reshape((naxis2_detector.value, naxis1_detector.value))
    else:
        if verbose:
            print('Skipping adding Gaussian readout noise')

    save_image2d_detector_method0(
        header_keys,
        image2d_detector_method0=image2d_detector_method0,
        prefix_intermediate_fits=prefix_intermediate_fits,
        bitpix=16,
    )

    # ---------------------------------------------------
    # compute image2d RSS from image in detector, method1
    # ---------------------------------------------------
    image2d_rss_method1 = compute_image2d_rss_from_detector_method1(
        image2d_detector_method0=image2d_detector_method0,
        naxis1_detector=naxis1_detector,
        naxis1_ifu=naxis1_ifu,
        nslices=nslices,
        dict_ifu2detector=dict_ifu2detector,
        wcs3d=wcs3d,
        noparallel_computation=noparallel_computation,
        verbose=verbose
    )

    # save FITS file
    save_image2d_rss(
        wcs3d=wcs3d,
        header_keys=header_keys,
        image2d_rss=image2d_rss_method1,
        method=1,
        prefix_intermediate_fits=prefix_intermediate_fits,
        bitpix=-32,
    )

    # ------------------------------------
    # compute image3d IFU from RSS method1
    # ------------------------------------
    image3d_ifu_method1 = compute_image3d_ifu_from_rss_method1(
        image2d_rss_method1=image2d_rss_method1,
        naxis1_detector=naxis1_detector,
        naxis2_ifu=naxis2_ifu,
        naxis1_ifu=naxis1_ifu,
        nslices=nslices,
        verbose=verbose
    )

    # save FITS file
    if len(prefix_intermediate_fits) > 0:
        hdu = fits.PrimaryHDU(image3d_ifu_method1.astype(np.float32))
        pos0 = len(hdu.header) - 1
        hdu.header.extend(wcs_to_header_using_cd_keywords(wcs3d), update=True)
        hdu.header.update(header_keys)
        hdu.header.insert(
            pos0, ('COMMENT', "FITS (Flexible Image Transport System) format is defined in 'Astronomy"))
        hdu.header.insert(
            pos0 + 1, ('COMMENT', "and Astrophysics', volume 376, page 359; bibcode: 2001A&A...376..359H"))
        hdul = fits.HDUList([hdu])
        outfile = f'{prefix_intermediate_fits}_ifu_3D_method1.fits'
        print(f'Saving file: {outfile}')
        hdul.writeto(f'{outfile}', overwrite='yes')



#
# Copyright 2024-2025 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#

"""Determine (X,Y) offsets between slices along NAXIS3"""

import argparse
from astropy.io import fits
import astropy.units as u
from astropy.wcs import WCS
import matplotlib.pyplot as plt
import numpy as np
from skimage.registration import phase_cross_correlation
import sys

from numina.array.display.polfit_residuals import polfit_residuals_with_sigma_rejection
from numina.array.display import shift_image2d
from numina.array.rescale_array_z1z2 import rescale_array_to_z1z2
from numina.array.yx_offsets_correlate2d import yx_offsets_correlate2d
from .compare_adr_extensions_in_3d_cube import compare_adr_extensions_in_3d_cube
from .ctext import ctext


def compute_i1i2(i, naxis3, binning_naxis3):
    """Auxiliary function to compute indices for binned slice

    Parameters
    ----------
    i : int
        Array index around which the desired interval [i1,i2] will be
        calculated.
    naxis3 : int
        Maximum allowed value.
    binning_naxis3 : float
        Bin width.

    Returns
    -------
    i1 : int
        Lower array index.
    i2 : int
        Upper array index.
    """

    i1 = int(i - binning_naxis3 / 2)
    if i1 < 0:
        i1 = 0
    i2 = int(i1 + binning_naxis3)
    if i2 > naxis3:
        i2 = naxis3
    return i1, i2


def measure_slice_xy_offsets_in_3d_cube(
        data3d,
        npoints_along_naxis3=100,
        polydeg=2,
        times_sigma_reject=3.0,
        islice_reference=None,
        iterate=True,
        method=1,
        plots=False,
        verbose=False
):
    """Compute (X, Y) offsets of the slices in a 3D datacube

    We are assuming that X correspond to NAXIS1, and Y to NAXIS2,
    where:

    NAXIS3, NAXIS2, NAXIS1 = data3d.shape

    Parameters
    ----------
    data3d : `~numpy.ndarray`
        Data cube.
    npoints_along_naxis3 : int
        Number of equidistant points along NAXIS3 for which the offsets
        in (X,Y) are calculated. In practice, a binning is performed
        along NAXIS3, where the number of slices to be summed in each
        bin is NAXIS3 / npoints_along_naxis3.
    polydeg : int
        Polynomial degree to fit distortion along NAXIS3.
    times_sigma_reject : float
        Times sigma to reject measured offsets when fitting the
        polynomial.
    islice_reference: int or None
        Pixel along NAXIS3 number to derive the initial reference
        binned slice. If None, the central location along NAXIS3 is
        employed. It must be a number from 1 to NAXIS3.
    iterate : bool
        If it's True, the procedure is repeated twice. In the first
        iteration, the binned slice centered at islice_refence is
        used as the reference slice. In the second iteration, the
        reference slice is obtained by averaging the entire corrected
        cube along NAXIS3.
    method : int
        Two methods are implemented:
        - method 1: using skimage.registration.phase_cross_correlation
        - method 2: using scipy.signal.correlate2d
        Both methods seem to yield similar results. Method 1 is faster
        and exhibits smaller residuals than method 2.
    plots : bool
        If True, plot intermediate results.
    verbose : bool
        If True, display intermediate information.

    Returns
    -------
    delta_x_array : `~numpy.ndarray`
        Offsets in X direction (pixels), evaluated along NAXIS3.
    delta_y_array : `~numpy.ndarray`
        Ofssets in Y direction (pixels), evaluated along NAXIS3.
    i1_ref : int
        First slice to derive the reference slice. The slice interval
        corresponds to [i1_ref:i2_ref] following the FITS convention.
    i2_ref : int
        Last slice to derive the reference slice. The slice interval
        corresponds to [i1_ref:i2_ref] following the FITS convention.

    """

    naxis3, naxis2, naxis1 = data3d.shape

    if npoints_along_naxis3 > naxis3:
        if verbose:
            print(f'WARNING: {npoints_along_naxis3=} > {naxis3=}')
            print('         forcing npoints_along_naxis3 = naxis3')
        npoints_along_naxis3 = naxis3

    # reference slice for the first iteration
    if islice_reference is None:
        islice_reference = naxis3 // 2
        if verbose:
            print(f'WARNING: setting {islice_reference=}')

    if islice_reference < 1:
        raise ValueError(f'Unexpected {islice_reference=} < 1')
    if islice_reference > naxis3:
        raise ValueError(f'Unexpected {islice_reference=} > {naxis3=}')

    # bin width along NAXIS3
    binning_naxis3 = naxis3 / npoints_along_naxis3
    if binning_naxis3 == 1:
        islice_array = np.arange(naxis3)
    else:
        islice_array = np.linspace(binning_naxis3 / 2, naxis3 - binning_naxis3 / 2, npoints_along_naxis3).astype(int)
    if verbose:
        print(f'{binning_naxis3=}')

    niterations = 1
    if iterate:
        niterations += 1

    # duplicate the input 3D array
    data3d_work = data3d.copy()

    # main loop in number of iteration
    delta_x_array = None
    delta_y_array = None
    i1_ref = None
    i2_ref = None
    for iteration in range(niterations):
        # compute reference binned slice
        if iteration == 0:
            i1_ref, i2_ref = compute_i1i2(islice_reference-1, naxis3, binning_naxis3)
            slice_reference = np.mean(data3d_work[i1_ref:i2_ref, :, :], axis=0)
        else:
            data3d_corrected = np.zeros_like(data3d_work)
            for islice in np.arange(naxis3):
                data3d_corrected[islice, :, :] = shift_image2d(
                    data3d_work[islice, :, :],
                    xoffset=-delta_x_array[islice],
                    yoffset=-delta_y_array[islice],
                    resampling=2
                )
            i1_ref = 0
            i2_ref = naxis3
            slice_reference = np.mean(data3d_corrected[i1_ref:i2_ref, :, :], axis=0)
            # del data3d_corrected
        if verbose:
            # FITS convention
            print(f'iteration: {iteration} --> reference slice in [{i1_ref+1}:[{i2_ref}]')
        if plots:
            fig, ax = plt.subplots()
            ax.imshow(slice_reference, origin='lower')
            ax.set_xlabel('Array index (NAXIS1 direction)')
            ax.set_ylabel('Array index (NAXIS2 direction)')
            # FITS convention
            ax.set_title(f'slice_reference along NAXIS3\n(i1_ref: {i1_ref+1}, i2_ref: {i2_ref}, iteration {iteration})')
            plt.tight_layout()
            plt.show()

        # rescale image to (0, 1)
        slice_reference, _ = rescale_array_to_z1z2(slice_reference, (0, 1))

        # loop in number of binned slice
        xfit = []
        delta_x_slice = []
        delta_y_slice = []
        for islice in islice_array:
            i1, i2 = compute_i1i2(islice, naxis3, binning_naxis3)
            xfit.append((i1 + i2 - 1) / 2)
            slice_binned = np.nanmean(data3d_work[i1:i2, :, :], axis=0)
            slice_binned, _ = rescale_array_to_z1z2(slice_binned, (0, 1))
            if method == 1:
                yx_offsets, _, _ = phase_cross_correlation(
                    reference_image=slice_reference,
                    moving_image=slice_binned,
                    upsample_factor=100,
                    overlap_ratio=0.90
                )
            elif method == 2:
                yx_offsets = yx_offsets_correlate2d(
                    reference_image=slice_reference,
                    moving_image=slice_binned
                )
            else:
                raise ValueError(f'Unexpected {method=}')
            delta_x_slice.append(-yx_offsets[1])
            delta_y_slice.append(-yx_offsets[0])
        xfit = np.array(xfit)
        delta_x_slice = np.array(delta_x_slice)
        delta_y_slice = np.array(delta_y_slice)
        for yfit, title in zip(
                [delta_x_slice, delta_y_slice],
                ['delta_x_slice', 'delta_y_slice']
        ):
            if plots:
                debugplot = 1
                fig = plt.figure()
            else:
                debugplot = 0
                fig = None
            poly, residuals, reject = polfit_residuals_with_sigma_rejection(
                x=xfit,
                y=yfit,
                deg=polydeg,
                times_sigma_reject=times_sigma_reject,
                title=title,
                debugplot=debugplot,
                fig=fig
            )
            if verbose:
                print(f'{title} -> {poly=}')
            if plots:
                plt.tight_layout()
                plt.show()
            result = poly(np.arange(naxis3))
            if title == 'delta_x_slice':
                delta_x_array = result
            else:
                delta_y_array = result

    return delta_x_array, delta_y_array, i1_ref+1, i2_ref   # FITS convention


def main(args=None):

    # parse command-line options
    parser = argparse.ArgumentParser(description="Determine (X,Y) offsets between slices along NAXIS3")
    parser.add_argument("filename", help="Input 3D FITS file")
    parser.add_argument("npoints", help="Number of points along NAXIS3", type=int)
    parser.add_argument("--extname", help="Output extension name to store result (default None)",
                        type=str, default='None')
    parser.add_argument("--polydeg", help="Polynomial degree to fit distortion", type=int, default=2)
    parser.add_argument("--times_sigma_reject",
                        help="Times sigma to reject fitted offsets (default 3.0)",
                        type=float, default=3.0)
    parser.add_argument("--islice_reference",
                        help="Initial pixel corresponding to the reference slice (from 1 to NAXIS3)",
                        type=int)
    parser.add_argument("--iterate", help="Force one iteration", action="store_true")
    parser.add_argument("--method", help="Method (1: skimage, 2: scipy)",
                        type=int, choices=[1, 2], default=1)
    parser.add_argument("--plots", help="Plot intermediate results", action="store_true")
    parser.add_argument("--verbose", help="Display intermediate information", action="store_true")
    parser.add_argument("--echo", help="Display full command line", action="store_true")

    args = parser.parse_args(args=args)

    if len(sys.argv) == 1:
        parser.print_usage()
        raise SystemExit()

    if args.verbose:
        for arg, value in vars(args).items():
            print(ctext(f'{arg}: {value}', faint=True))

    if args.echo:
        print('\033[1m\033[31mExecuting: ' + ' '.join(sys.argv) + '\033[0m\n')

    # protections
    extname = args.extname.upper()
    if len(extname) > 8:
        raise ValueError(f"Extension '{extname}' must be less than 9 characters")

    with fits.open(args.filename) as hdul:
        primary_header = hdul[0].header
        data3d = hdul[0].data

    if primary_header['NAXIS'] != 3:
        raise ValueError("Expected NAXIS=3 not found in PRIMARY HDU")

    delta_x_array, delta_y_array, i1_ref, i2_ref = measure_slice_xy_offsets_in_3d_cube(
        data3d=data3d,
        npoints_along_naxis3=args.npoints,
        polydeg=args.polydeg,
        times_sigma_reject=args.times_sigma_reject,
        islice_reference=args.islice_reference,
        iterate=args.iterate,
        method=args.method,
        plots=args.plots,
        verbose=args.verbose
    )

    # save result in extension
    if extname != 'NONE':
        if args.verbose:
            print(f'Updating file {args.filename}')
        # binary table to store result
        col1 = fits.Column(name='Delta_x', format='D', array=delta_x_array, unit='pixel')
        col2 = fits.Column(name='Delta_y', format='D', array=delta_y_array, unit='pixel')
        hdu_result = fits.BinTableHDU.from_columns([col1, col2])
        hdu_result.name = extname.upper()
        hdu_result.header['METHOD'] = (args.method, '1: skimage, 2: scipy')
        hdu_result.header['POLYDEG'] = (args.polydeg, 'Polynomial degree to fit distortion')
        hdu_result.header['TSIGMA'] = (args.times_sigma_reject, 'Times sigma to reject fitted offsets')
        hdu_result.header['NP_FIT'] = (args.npoints, 'Number of points along NAXIS3')
        hdu_result.header['ITERATE'] = (args.iterate, 'Iterate procedure (T: True, F: False)')
        if args.islice_reference is not None:
            i0_ref = args.islice_reference
        else:
            i0_ref = -1
        hdu_result.header['I0_REF'] = (i0_ref, 'Initial pixel for reference slice (-1: None)')
        hdu_result.header['I1_REF'] = (i1_ref, 'First pixel of reference slice')
        hdu_result.header['I2_REF'] = (i2_ref, 'Last pixel of reference slice')
        wcs3d = WCS(primary_header)
        naxis1, naxis2, naxis3 = wcs3d.pixel_shape
        wave = wcs3d.spectral.pixel_to_world(np.arange(naxis3))
        reference_vacuum_wavelength = (wave[i1_ref-1] + wave[i2_ref-1]) / 2
        hdu_result.header['REFEWAVE'] = (reference_vacuum_wavelength.to(u.m).value,
                                         'Reference vacuum wavelength (m)')
        # open and update existing FITS file
        hdul = fits.open(args.filename, mode='update')
        if extname in hdul:
            if args.verbose:
                print(f"Updating extension '{extname}'")
            hdul[extname] = hdu_result
        else:
            if args.verbose:
                print(f"Adding new extension '{extname}'")
            hdul.append(hdu_result)
        hdul.flush()
        hdul.close()

    # display results
    if args.plots:
        suptitle = f"file: {args.filename} (extension: {extname})\nreference slices in [{i1_ref}:{i2_ref}]"
        compare_adr_extensions_in_3d_cube(args.filename, extname1=extname, extname2=None, suptitle=suptitle)


if __name__ == "__main__":

    main()




"""Side effect decorator

Decorates a function to call another
"""


class FuncCall(object):
    """Record a function call"""

    def __init__(self):
        self.clear()

    def __call__(self, *args, **kwargs):
        self.called = True
        self.args = args
        self.kwargs = kwargs

    def clear(self):
        self.called = False
        self.args = ()
        self.kwargs = {}


def side_effect(effect_func):
    import functools

    def side_effect_decorator(func):
        @functools.wraps(func)
        def func_wrapper(*args, **kwargs):
            try:
                effect_func(*args, **kwargs)
            except Exception:
                # ignoring side effect exceptions
                pass
            return func(*args, **kwargs)

        func_wrapper.side_effect = effect_func

        return func_wrapper
    return side_effect_decorator


def record_call(func):

    rcd = side_effect(FuncCall())
    return rcd(func)



#
# Copyright 2020 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#

"""Build a LoadableDRP from a yaml file"""

import numina.core.dataload


load = numina.core.dataload.DataLoaders()


@load.register('image/fits', priority=20)
def load_fits_0(pathname):
    import astropy.io.fits as fits
    return fits.open(pathname)


@load.register('application/json', priority=20)
def load_json(pathname):
    import json
    with open(pathname) as fd:
        return json.load(fd)


@load.register('application/json', numina.core.dataload.is_json_structured, priority=5)
def load_json(pathname):  # noqa: F811
    import json
    from numina.util.objimport import import_object

    with open(pathname) as fd:
        data = json.load(fd)
    type_fqn = data['type_fqn']
    cls = import_object(type_fqn)
    obj = cls.__new__(cls)
    obj.__setstate__(data)
    return obj


describe = numina.core.dataload.DataLoaders()


check = numina.core.dataload.DataChecker()


# Here we could have methods to extract
# this information from different files
# FITS and JSON
_describe_keys = [
    'instrument',
    'object',
    'observation_date',
    'uuid',
    'type',
    'mode',
    'exptime',
    'darktime',
    'insconf',
    'blckuuid',
    'quality_control'
]


@describe.register('image/fits', priority=20)
def describe_fits_0(pathname):
    import astropy.io.fits as fits
    with fits.open(pathname) as hdulist:
        prim = hdulist[0].header
        instrument = prim.get("INSTRUME", "unknown")
        obsmode = prim.get("OBSMODE", "unknown")

        return instrument, obsmode


@describe.register('application/json',
                   numina.core.dataload.is_json_structured,
                   priority=20)
def describe_json(pathname):
    import json
    from numina.util.objimport import import_object

    with open(pathname) as fd:
        data = json.load(fd)

    type_fqn = data['type_fqn']
    cls = import_object(type_fqn)
    obj = cls.__new__(cls)
    obj.__setstate__(data)
    return obj.instrument, "TBD"



#
# Copyright 2008-2021 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#

"""
Recipe requirement holders
"""


import numina.exceptions
import numina.types.obsresult as obtypes

from .dataholders import Requirement


class ObservationResultRequirement(Requirement):
    """The Recipe requires the result of an observation."""

    def __init__(self, query_opts=None):

        super(ObservationResultRequirement, self).__init__(
            obtypes.ObservationResultType, "Observation Result",
            query_opts=query_opts
        )

    def __repr__(self):
        sclass = type(self).__name__
        fmt = "%s(dest=%r, description='%s')"
        msg = fmt % (sclass, self.dest, self.description)
        return msg

    def query(self, dal, obsres, options=None):
        """

        Parameters
        ----------
        dal
        obsres
        options

        Returns
        -------

        """
        from numina.core.query import ResultOf

        q_options = self.query_options(options)

        if isinstance(q_options, ResultOf):
            dest_field = 'frames'
            dest_type = list
            # Field to insert the results
            if not hasattr(obsres, dest_field):
                setattr(obsres, dest_field, dest_type())

            dest_obj = getattr(obsres, dest_field)

            values = dal.search_result_relative(self.dest, self.type, obsres,
                                                result_desc=q_options)

            for partial in values:
                dest_obj.append(partial.content)

        return obsres

    def on_query_not_found_from_type(self, notfound):
        """

        Parameters
        ----------
        notfound

        Returns
        -------

        """
        raise numina.exceptions.NoResultFound(
            'unable to complete ObservationResult') from notfound




import pkgutil

import numina.drps
import numina.drps.drpbase
import numina.core.pipelineload as pload

from ..cli import main


drpdata1 = pkgutil.get_data('numina.drps.tests', 'drptest1.yaml')
drpdata2 = pkgutil.get_data('numina.drps.tests', 'drptest2.yaml')


def test_show_modes(capsys, monkeypatch):
    """Test that one instrument is shown"""

    def mockreturn():
        import numina.drps.drpbase
        import numina.core.pipelineload as pload
        drps = {}

        drp1 = pload.drp_load_data('numina', drpdata1)
        drps[drp1.name] = drp1
        return numina.drps.drpbase.DrpGeneric(drps)

    monkeypatch.setattr(numina.drps, "get_system_drps", mockreturn)

    expected = ["Observing Mode: 'Fail' (fail)",
                " summary: Summary of fail recipe",
                " instrument: TEST1",
                "Observing Mode: 'Bias' (bias)",
                " summary: Summary of Bias recipe",
                " instrument: TEST1",
                "Observing Mode: 'Dark' (dark)",
                " summary: Summary of Dark recipe",
                " instrument: TEST1",
                "Observing Mode: 'Image' (image)",
                " summary: Summary of Image recipe",
                " instrument: TEST1",
                ""
                ]

    main(['show-modes'])

    out, err = capsys.readouterr()
    out = out.split("\n")
    out.sort()
    expected.sort()
    assert out == expected


def test_show_2_instruments(capsys, monkeypatch):
    """Test that two instruments are shown"""

    def mockreturn():
        drps = {}

        drp1 = pload.drp_load_data('numina', drpdata1)
        drp2 = pload.drp_load_data('numina', drpdata2)
        drps[drp1.name] = drp1
        drps[drp2.name] = drp2
        return numina.drps.drpbase.DrpGeneric(drps)

    monkeypatch.setattr(numina.drps, "get_system_drps", mockreturn)

    expected = ['',
                "Observing Mode: 'Fail' (fail)",
                ' summary: Summary of fail recipe',
                ' instrument: TEST1',
                "Observing Mode: 'Bias' (bias)",
                ' summary: Summary of Bias recipe',
                ' instrument: TEST1',
                "Observing Mode: 'Dark' (dark)",
                ' summary: Summary of Dark recipe',
                ' instrument: TEST1',
                "Observing Mode: 'Image' (image)",
                ' summary: Summary of Image recipe',
                ' instrument: TEST1',
                "Observing Mode: 'Success' (success)",
                ' instrument: TEST2',
                ' summary: Summary of success recipe',
                "Observing Mode: 'Dark' (dark)",
                ' summary: Dark recipe',
                ' instrument: TEST2',
                ]

    main(['show-modes'])

    out, err = capsys.readouterr()
    out = out.split("\n")
    out.sort()
    expected.sort()
    assert out == expected


def test_show_modes_no_instruments(capsys, monkeypatch):
    """Test that no instruments are shown"""

    def mockreturn():
        return numina.drps.drpbase.DrpGeneric()

    monkeypatch.setattr(numina.drps, "get_system_drps", mockreturn)

    expected = ""

    main(['show-modes'])

    out, err = capsys.readouterr()

    assert out == expected


def test_show_modes_2_instruments_select_no(capsys, monkeypatch):
    """Test that two instruments are shown"""

    def mockreturn():
        drps = {}
        drp1 = pload.drp_load_data('numina', drpdata1)
        drp2 = pload.drp_load_data('numina', drpdata2)
        drps[drp1.name] = drp1
        drps[drp2.name] = drp2
        return numina.drps.drpbase.DrpGeneric(drps)

    monkeypatch.setattr(numina.drps, "get_system_drps", mockreturn)

    expected = ["No instrument named: TEST3",
                ""
                ]

    main(['show-modes', '-i', 'TEST3'])

    out, err = capsys.readouterr()
    out = out.split("\n")
    out.sort()
    expected.sort()
    assert out == expected


def test_show_modes_no_instruments_select_no(capsys, monkeypatch):
    """Test that two instruments are shown"""

    def mockreturn():
        drps = {}
        return numina.drps.drpbase.DrpGeneric(drps)

    monkeypatch.setattr(numina.drps, "get_system_drps", mockreturn)

    expected = ["No instrument named: TEST3",
                ""
                ]

    main(['show-modes', '-i', 'TEST3'])

    out, err = capsys.readouterr()
    out = out.split("\n")
    out.sort()
    expected.sort()
    assert out == expected



#
# Copyright 2008-2017 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#

"""Metaclasses for Recipes."""

from .recipeinout import RecipeResult, RecipeInput
from .dataholders import EntryHolder, Result
from .requirements import Requirement

_RECIPE_RESULT_NAME = 'RecipeResult'
_RECIPE_INPUT_NAME = 'RecipeInput'


class RecipeType(type):
    """Metaclass for Recipe."""
    def __new__(cls, classname, parents, attributes):

        filter_reqs = {}
        filter_prods = {}
        filter_attr = {}

        for name, val in attributes.items():
            if isinstance(val, EntryHolder):
                if isinstance(val, Requirement):
                    filter_reqs[name] = val
                if isinstance(val, Result):
                    filter_prods[name] = val
            else:
                filter_attr[name] = val

        BaseRecipeResult = cls.get_base_class(
            RecipeResult, parents, attributes, _RECIPE_RESULT_NAME)
        BaseRecipeInput = cls.get_base_class(
            RecipeInput, parents, attributes, _RECIPE_INPUT_NAME)

        ReqsClass = cls.create_inpt_class(
            classname, BaseRecipeInput, filter_reqs)
        ResultClass = cls.create_prod_class(
            classname, BaseRecipeResult, filter_prods)

        filter_attr[_RECIPE_RESULT_NAME] = ResultClass
        filter_attr[_RECIPE_INPUT_NAME] = ReqsClass
        filter_attr[ResultClass.__name__] = ResultClass
        filter_attr[ReqsClass.__name__] = ReqsClass

        return super(RecipeType, cls).__new__(
            cls, classname, parents, filter_attr)

    @classmethod
    def create_gen_class(cls, classname, baseclass, attributes):
        if attributes:
            klass = type(classname, (baseclass,), attributes)
            # Add docs
            klass = generate_docs(klass)
        else:
            klass = baseclass
        return klass

    @classmethod
    def get_base_class(cls, default, parents, attributes, name):
        # Find base class for RecipeResult in parents
        if name in attributes:
            return attributes[name]
        else:
            for parent in parents:
                if hasattr(parent, name):
                    return getattr(parent, name)
            else:
                return default

    @classmethod
    def create_inpt_class(cls, classname, base, attributes):
        return cls.create_gen_class(f'{classname}Input',
                                    base, attributes)

    @classmethod
    def create_prod_class(cls, classname, base, attributes):
        return cls.create_gen_class(f'{classname}Result',
                                    base, attributes)


def generate_docs(klass):
    """Add documentation to generated classes"""
    import numina.types.datatype

    attrh = ('Attributes\n'
             '----------\n')

    doc = getattr(klass, '__doc__', None)

    if doc is None or doc == '':
        doc = f"{klass.__name__} documentation."

    if len(klass.stored()):
        doc = doc + '\n\n' + attrh

    skeys = sorted(klass.stored().keys())
    for key in skeys:
        y = klass.stored()[key]

        if isinstance(y, Requirement):
            modo = 'requirement'
        elif isinstance(y, Result):
            modo = 'product'
        else:
            modo = ""
        if y.type.isproduct():
            tipo = y.type.__class__.__name__
        elif isinstance(y.type, numina.types.datatype.PlainPythonType):
            tipo = y.type.internal_type.__name__
        else:
            tipo = y.type.__class__.__name__

        if y.optional:
            if y.default_value():
                modo = f"{modo}, optional, default={y.default}"
            else:
                modo = f"{modo}, optional"

        descript = y.description
        if descript:
            field = f"{key} : {tipo}, {modo}\n {descript}\n"
        else:
            field = f"{key} : {tipo}, {modo}\n"
        doc = doc + field

    klass.__doc__ = doc
    return klass
